/*
Module : FileTreeCtrl.cpp
Purpose: Implementation for an MFC class which provides a tree control similiar 
         to the left hand side of explorer
Created: PJN / 25-12-1999
History: PJN / 11-01-2000 1. Added some asserts to HasGotSubEntries
                          2. Fixed a problem with calling OnDblclk when no tree item is selected
                          3. Removed an unused variable from SetSelectedPath
         PJN / 25-01-2000 1. Minor update to the code in CFileTreeCtrl::OnDblclk to only allow 
                          selected items to be opened.
         PJN / 31-01-2000 1. Fixed a problem when you right mouse click over a non - selected item.
                          The control now implements the same behavior as Explorer for this.
                          2. Removed check for keyboard invocation of the context menu in OnContextMenu
                          3. Now displays the context menu over the selected item when invoked via
                          the keyboard. Again this is the same bahavior as Explorer has.
                          4. Optimized the code in PreTranslateMessage
                          5. Fixed a bug in CFileTreeCtrl::OnEndlabeledit
         PJN / 02-04-2000 1. Fixed a small bug in CFileTreeCtrl::SetRootFolder
                          2. Fixed the problem with initialisation errors in the code. Client code must not 
                          explicitly call PopulateTree when the window is created. When used in a dialog
                          resource this is not necessary as it is called for you in the DDX functions.
         PJN / 13-05-2000 1. Fixed a problem where items on the context menu were being invoked for the 
                          wrong item when you right mouse click over an item that is not the selected item.
                          Behaviour now is that the item is selected prior to showing the context menu. 
                          Now, this is same behaviour as Explorer.
         PJN / 12-07-2000 1. Now uses ON_NOTIFY_REFLECT_EX instead of ON_NOTIFY_REFLECT for handling reflected
                          messages. This allows derived classes to handle these messages also. Thanks to 
                          Christian Dahl for this.
                          2. Sample app now allows drag drop behaviour to be toggled
                          3. Fixed a problem whereby two items were left selected after you did a drag /
                          drop operation. Thanks to Jonathon Ralston for this.
                          4. Removed function declaration for unused function "InsertDriveItem".
                          5. Removed an unreferenced variable in InsertFileItem.
                          6. Tidied up the UpOneLevel functions and made it public.
                          7. Removed all the message handlers in the sample code which reflect straight
                          down to the tree control. Instead the OnCmdMsg first routes the message to this
                          class.
                          8. Renamed all menu items which CFileTreeCtrl uses to include the prefix TREEFILECTRL
                          9. Renamed all the methods to more generic names
                          10. PreTranslateMessage now uses PostMessage instead of calling functions directly. 
                          This allows up to function correctly for derived classes in addition to correctly
                          disabling items through the OnUpdate mechanism
                          11. Removed an unreferrenced variable in OnRclick
                          12. Removed the unreferrenced variable m_hSelItem
                          13. Optimized a number of expressions by putting the boolean comparisons first
                          14. m_bAllowRename parameter is now observed for in place editing of an item
                          15. Now supports hiding of Drive types via the SetDriveHideFlags function. See the 
                          menu options on the Tools menu in the sample program for its usage.
                          16. Filename masks can now be specified via the SetFileNameMask method. See the 
                          menu options on the Tools menu in the sample program for its usage.
                          17. File types can now be specified via the GetFileHideFlags function. See the 
                          menu options on the Tools menu in the sample program for its usage.
                          18. Fixed a small issue in one of my calls to ::GetKeyState
                          19. Fixed a bug where programs was crashing if an icon index for it could not
                          be found.
                          20. Made many of the methods of CFileTreeCtrl virtual, thus allowable better
                          use in end user derived classes.
                          21. Fixed problem where SetSelectedPath(_T("C:\\"), FALSE) was resulting 
                          in the drive being expanded even through FALSE was being sent in to specify
                          that the item should not be expanded.
                          22. A virtual "CanDisplayFile" has been added to allow you to decide at runtime
                          whether or not a certain file is to be displayed.
                          23. A virtual "CanDisplayFolder" has been added to allow you to decide at
                          runtime whether or not a certain folder is to be displayed
                          24. Now optionally displays compressed files in a different color, similiar to
                          explorer. The color is customizable through the class API.
                          25. Code has been made smarter so that it does not have to spin up the floppy
                          disk to determine if there are files on it. It now initially displays a "+"
                          and only when you try to expand it will it do the actual scan.
         PJN / 23-07-2000 1. Fixed a bug where the expansion state of the selected item was not being
                          preserved across refreshes.
                          2. Now includes full support for Next / Prev support similiar to Windows 
                          Explorer with the Desktop Update.
                          3. Updated sample app to have some useful toolbars.
                          4. Changing any tree settings which can affect its appearance now force
                          a refresh of its contents.
                          5. ItemToPath method has been made const.
                          6. Addition of PathToItem method
                          7. Auto refresh of items is now provided for by means of change notification
                          threads. This is configurable via the SetAutoRefresh method.
                          8. The root folder of the tree control can now be changed from the sample app
                          9. Fixed a bug in SetRootFolder when passed an empty folder
                          10. Fixed a bug where the system image list was not being initialized correctly
                          if the user did not have a "C:\\" drive. This could occur on NT/Windows2000
                          11. Fixed a bug in IsFile and IsFolder which was causing invalid files or folders
                          to appear valid.
                          12. Deleted items are now removed upon expansion. Also if the item being expanded was
                          deleted and it was the only child, then its parent has the "-" expansion button removed
                          13. Removable drive nodes are collapsed back to their root nodes if their media is 
                          changed in the intervening time when a node expansion occurs.
                          14. Wait cursor is now displayed while a refresh is taking place.
                          15. A "OnSelectionChanged" virtual function has now been provided
                          16. Sample app's icon has been made the same as Explorers.
                          17. Sample app now displays the path name of the currently selected item in the tree control.
                          18. Fixed a bug in IsCompressed
                          19. items are now deleted when selected if they do not exist.
         PJN / 05-09-2000 1. Fixed a bug in CFileTreeCtrl::IsFile and CFileTreeCtrl::IsFolder
         PJN / 20-09-2000 1. Control now includes DECLARE_DYNCREATE thereby allowing it to be used
                          in code which requires this such as Stingray's SEC3DTabWnd. Thanks to Petter Nilsen for
                          pointing out this omission
         PJN / 02-10-2000 1. Fixed a stack overwrite problem in CSystemImageList::CSystemImageList
                          2. Removed an unreferrenced variable in CFileTreeCtrl::OnSelChanged
                          3. Removed an unreferrenced variable in CFileTreeCtrl::OnItemExpanding
                          4. Changed the SendMessage in CFileTreeCtrl::OnDblClk to prevent a crash 
                          which was occurring when the open call sometimes caused a refresh call 
                          which changed the tree items at times. When the double click message handler
                          continued it then caused item expand notifications for items already deleted
                          and of course crashes.
                          5. Removed an unreferrenced variable in CFileTreeCtrl::EndDragging
                          6. Removed an unreferrenced variable in CFileTreeCtrl::KillNotificationThread
                          7. Sample app now remembers the selected path and its expansion state across
                          invocations.
         PJN / 05-05-2001 1. Updated copyright message.
                          2. Fixed a resource leak where icon resources were not being released. Thanks to Jay Kohler for
                          spotting this problem
         PJN / 05-08-2001 1. You can now optionally display Network Neighborhood
                          2. You can now turn on / off display of file extensions.
                          3. You can now display shared items with a different icon
                          4. Friendly names can now be displayed for drives.
         PJN / 11-08-2001 1. Improved checking to see if action is allowed in Rename and Delete
                          2. Fixed a bug in OnBeginLabelEdit
                          3. Fixed a problem in OnEndLabelEdit which was causing renames to fail when filename extensions 
                          were not being shown.
         PJN / 11-08-2001 1. Fixed a bug in OnSelChanged which was causing a crash when you right click on an empty area of the control.
                          Thanks to Eli Fenton for spotting this one.
                          2. The control now by default shows drives as children of "My Computer" just like in Explorer.
                          3. When you display a rooted directory in the control, you now have the option of displaying the root
                          folder in the control as the root item.  Thanks to Eli Fenton for suggesting this.
         PJN / 26-10-2001 1. Fixed some stability problems with the code. This was due to adding items to the system image list.
                          This is normally a very bad thing. Instead now the code uses TreeView custom draw (just like the blue color
                          for compressed items) to draw the icons for the Network Neighborhood items. Thanks to Darken Screamer and
                          Timo Haberkern for spotting this problem.
         PJN / 24-12-2001 1. Fixed a copy and paste bug in GoForward. Thanks to Michael T. Luongo for this fix.
                          2. Now allows encrypted files to be displayed in a different color
                          3. Fixed memory leak which was occurring when control was being used in a dialog
                          4. Fixed a problem with the calculation of indents when the style "TVS_LINESATROOT" is used.
         PJN / 16-02-2002 1. Updated copyright message
                          2. Fixed a drag/drop problem which caused the tree state to be inconsistent after the file was dropped.
                          3. Fixed a bug in the refresh code which was causing it to not reselect the selected node
                          after the refresh occurs. Thanks to John Nol for this fix.
                          4. Fixed a problem where the custom draw icons for network nodes were not being drawn in the correct
                          positions when scrollbars were present in the control. Again thanks to John Nol for this fix.
                          5. Fixed a bug in SetSelectedPath which would not display the correct selection if the node we 
                          want to select has been deleted due to the node becoming deleted when it was previously collapsed.
                          Thanks to Franz Fackelmann and John Nol for spotting this problem.
         PJN / 05-06-2002 1. Implemented function "SetUsingDifferentColorForEncrypted" which was declared but had no
                          implementation.
                          2. Fixed report of uninitialized member variable "m_nTimerID". Thanks to Emil Isberg for spotting this.
         PJN / 07-08-2002 1. Fixed a bug in the sample app which ships with the class which was occuring when you compiled
                          the code in Visual Studio.Net. This was due to MS changing the location oleimpl2.h header file. Thanks
                          to Darren Schroeder for spotting this problem.
         PJN / 22-09-2002 1. Removed a number of unreferrenced variables from the code, as highlighted by Visual Studio.Net. Thanks
                          to Bill Johnson for spotting this.
         PJN / 31-12-2002 1. Items are now sorted using the TVM_SORTCHILDRENCB Tree Control message. This means that the class
                          is no longer dependent on the CSortedArray class. That class has been left in this download for 
                          convenience only. Thanks to "Vladi" for this suggestion. 
         PJN / 26-02-2003 1. Fixed a bug in CFileTreeCtrl::PathToItem which was causing refreshes to fail when a rooted folder
                          was being displayed. Thanks to Michael J. Ryan for spotting this.
         PJN / 26-02-2003 1. Fixed more problems with refreshing and rooted paths. Thanks to Michael J. Ryan for spotting this.
         PJN / 31-05-2003 1. Fixed a bug where the filename mask was being ignored.
         PJN / 10-06-2003 1. Custom draw is now not used for items underneath Network Neighborhood. This means that some items there will
                          just display the standard network neighborhood icon. To resolve this issue you would need to implement
                          a completely new solution using the shell API's. Thanks to John Nol for reporting a custom draw bug in this
                          area which has forced this change.
                          2. m_sFQPath is now not filled in for virtual items such as My Computer and Network Neighborhood. Thanks
                          to Hans Dietrich for this suggestion.
         PJN / 05-07-2003 1. Updated the WM_DEVICECHANGE message handler in the sample app to use the correct function signature for
                          Visual Studio 6 and later. Thanks to Ed Eichman for reporting this issue.
                          2. Fixed a bug in the documentation of CFileTreeCtrl::SetDriveHideFlags. Thanks to Ed Eichman for 
                          reporting this issue.
                          3. Now documents the functions SetAutoRefresh and GetAutoRefresh. In addition by default auto refresh is now
                          turned off. Thanks to Ed Eichman for pointing this out.
         PJN / 13-01-2004 1. Used a new form of pragma pack to avoid resetting the packing size to the default when it is currently
                          set to a non standard value.
         PJN / 09-07-2004 1. Fixed a memory leak in the NETRESOURCE structure if client code returned FALSE from 
                          CFileTreeCtrl::CanDisplayNetworkItem. Thanks to Michael Oerder reporting this problem.
         PJN / 17-04-2005 1. Addition of TVIF_HANDLE to the mask fields in the functions GetIconIndex(HTREEITEM hItem), 
                          GetSelIconIndex(HTREEITEM hItem) and SetHasPlusButton(HTREEITEM hItem, BOOL bHavePlus). This avoids ASSERTs
                          when the class is derived from the SECTreeCtrl class in the Stingray Objective Toolkit. Thanks to Greg Marr
                          for reporting this problem.
                          2. Addition of a "FILETREECTRL_EXT_CLASS" define to the class defintions. This makes the code easier to use
                          in an extension DLL.
                          3. Code now uses the TVITEM structure name rather than TV_ITEM which is in line with what is recommended by MS.
                          4. Code now uses the TVINSERTSTRUCT structure name rather than TV_INSERTSTRUCT which is in line with what is 
                          recommended by MS.
                          5. Class now uses a preprocessor define of FILETREECTRL_BASE_CLASS to define the class which CFileTreeCtrl
                          is derived from. This allows the class to be used with the Stingray Objective Toolkit without needing to 
                          modify the CFileTreeCtrl source code.
                          6. Optimized the code which deletes child nodes by now using TVE_COLLAPSERESET flag.
                          7. Removed the unused function NumberOfChildItems.
                          8. Code now uses the NMTREEVIEW structure name rather than NM_TREEVIEW which is in line with what is 
                          recommended by MS.
                          9. Removed the now unused DeleteChildren method.
         PJN / 21-06-2005 1. Fixed a bug where drag and drag mechanism had problems when trying to drag an item into a folder immediately
                          after the root item. Thanks to Larry Paul for reporting this bug and provinding the fix.
                          2. Fixed a number of compiler warnings when the code is compiled using Visual Studio .NET 2003
         PJN / 25-01-2006 1. Updated copyright details.
                          2. Removed the GetChecked / SetChecked methods as CTreeCtrl as of VC++ 6 includes equivalent functionality. 
                          3. The CanDisplayFolder method has been updated to use logic similar to the CanDisplayFile virtual function.
                          This new functionality is controlled by a new m_dwFolderHideFlags variable which can be modified via the 
                          functions SetFolderHideFlags & GetFolderHideFlags. Thanks to Bob O'Neil for this update.
                          4. Reviewed all TRACE statements for correctness.
                          5. Fixed an issue where you get an ASSERT in the code when you are using XP manifest support and you are using
                          the TVS_CHECKBOXES style in your CFileTreeCtrl. The issue seems to be a bug in Commctrl v6 custom draw 
                          notifications as this problem does not occur if you are not using XP themes. Thanks to Bob O'Neill for this update.
         PJN / 18-06-2006 1. Replaced all calls to ZeroMemory with memset
                          2. Replaced all calls to CopyMemory with memcpy
                          3. Optimized CFileTreeCtrlItemInfo constructor code
                          4. Now includes code to optionally show "Desktop" as the root node. Thanks to Hans Dietrich for this very nice 
                          addition
                          5. Removed some now deleted files from the project.
                          6. Fixed a minor typo in the Allow Properties menu item in the sample app.
                          7. Now includes code to optionally show "My Documents". Thanks to Hans Dietrich for this very nice addition.
                          8. Fixed bug where m_bExtensionHidden setting was not being copied in CFileTreeCtrlItemInfo constructor
                          9. Renamed SetDisplayNetwork to SetShowNetwork and GetDisplayNetwork to GetShowNetwork for consistency with
                          all the other functions of the same type.
         PJN / 22-06-2006 1. Fixed a bug where expansion of the Desktop node fails on subsequent expansions. Thanks to Jan Moorman for
                          reporting this issue.
         PJN / 01-07-2006 1. Fixed a bug where expansion of UNC folders failed. Thanks to Jan Moorman for reporting this bug.
         PJN / 01-07-2006 1. Updated the code to clean compile on VC 2005     
                          2. Code now uses new C++ style casts rather than old style C casts where necessary.   
         PJN / 17-03-2007 1. Updated copyright details.  
                          2. Updated the sample apps to clean compile on VC 2005
                          3. Added a preprocessor define "FILETREECTRL_NO_RESOURCES" which allows you to use the CFileTreeCtrl code without
                          the need to pull in any resources. Please note that this will lead to reduced functionality of the control, for
                          example drag and drag will not be supported and all the menu functionality will also be excluded. For exact
                          details on what is out when you define this value, just look for all occurences of FILETREECTRL_NO_RESOURCES
                          in the FileTreeCtrl.cpp module. Thanks to saukwan chou for reporting this issue.
                          4. Optimized CFileTreeCtrl constructor code.
                          5. Optimized CFileTreeCtrlThreadInfo constructor code.
                          6. Optimized the code which gets the icon and "selected" icon for network servers which are displayed by the class.
                          What the code does is cache the local machine name in the constructor and uses it to get the icon for a server
                          when the code is enumerating servers. This does mean that if MS ever brings in a mechanism to allow customization
                          of the icon on a server by server level, then CFileTreeCtrl will ignore this. Thanks to Fotis Xomeritakis for
                          providing this optimization which really helps speed up the performance of CFileTreeCtrl when browsing domains or 
                          workgroups with a large number of workstations and servers.
                          7. Optimized CShareEnumerator constructor code.
         PJN / 31-03-2007 1. The icons used for computer nodes in Network neighborhood is now cached in the CFileTreeCtrl constructor. This is a 
                          further improvement over the caching of the local computer name approach which was used in the last version of the
                          code. If you do not want to use this approach, then you can define a FILETREECTRL_CACHE_COMPUTER_ICON preprocessor
                          value which will result in the computer icon being found when actual network enumeration is performed. Thanks to
                          Fotis Xomeritakis for this nice optimization.
         PJN / 09-04-2007 1. Updated IsDropSource method to now allow drag / drop of folders. Thanks to Kirk A. Baker for reporting this 
                          useful update.
                          2. The class now supports multiple filename masks. To support this new functionality, an updated / new set of 
                          functions have been provided, namely: AddMultipleFileNameMasks, SetFileNameMask, AddFileNameMask, RemoveFileNameMask, 
                          GetFileNameMask & GetFileNameMasks. Thanks to Fedor Gavrilov for this nice update.
         PJN / 22-04-2007 1. Fixed a bug in CFileTreeCtrl::OnItemExpanding when the control is created with the TVS_SINGLEEXPAND style. What
                          was happening was that if for example you selected C:\program files using SetSelectedPath, then you selected "C:\"
                          using the same function and then you selected "D:\", the control will end up showing no plus button on the C node.
                          Thanks to John Emmas for reporting this quite difficult to find bug.
         PJN / 06-06-2007 1. Now supports in place creation of folders. Thanks to Phil Gibson for prompting this update.
         PJN / 09-06-2007 1. Updated comments in documentation about usage of the Platform SDK.
                          2. Internal WM_TREEUPDATE_CHANGE message is now defined as WM_APP instead of WM_USER
                          3. CreateFolder method now takes a second parameter which allows the new child folder name to be set. If you leave 
                          this parameter as the default value of _T("") then the new folder name will be created by calling the virtual 
                          CreateNewChildFolderName function. This function has also been updated in this release to better encapsulate how a new
                          folder name is created.
                          4. Return value from CreateFolder method is now a HTREEITEM instead of a BOOL
                          5. Reworked the internal logic of CreateFolder method to avoid the need to call the Refresh method. As a result the
                          user experience is much improved due to the elimination of the tree updates.
         PJN / 13-06-2007 1. Reviewed class code for use of C++ casts instead of old style C casts.
                          2. Desktop node item now includes a file system path in its CFileTreeCtrlItemInfo pointer
                          3. Made all BOOL members of CFileTreeCtrlItemInfo a "bool" to conserve memory usage
                          4. Replaced CFileTreeCtrlItemInfo member variables "m_bDesktopNode", "m_bNetworkNode" & "m_bMyDocumentNode" with a 
                          single enum member variable
                          5. Fixed a bug where rename would previously be allowed for nodes such as My Documents, My Computers etc
                          6. Fixed a bug where delete would previously be allowed for nodes such as My Documents, My Computers etc
                          7. Desktop node now also includes files and folders as found on the desktop. Also the ordering of items on the 
                          desktop is consistent with how Explorer displays the desktop.
                          8. The shell is notified that a new directory has been created in CreateFolder via the SHChangeNotify API
         PJN / 16-10-2007 1. Fixed a bug where if you attempt to create a new folder within a drive and then immediately rename that folder 
                          to something else, the shell throws an error about renaming a file and changing it's parent drive. Thanks to 
                          Michael Stephenson for reporting this bug.
                          2. Fixed a bug where if you delete a folder and it is the only folder under it's parent folder, the parent retains it's 
                          "minus" icon, even though it no longer has any children. Thanks to Michael Stephenson for reporting this bug.
                          3. Fixed a bug where if you create a folder in a parent folder which did not have any child items, the parent folder
                          would not show the "minus" icon correctly. Thanks to Michael Stephenson for reporting this bug.
         PJN / 02-05-2008 1. Updated copyright details.
                          2. Updated the sample app to show how to implement a Vista style explorer tree view if it is running on Vista. This
                          functionality is implemented at the client app level instead of the control code itself, so you can use this code
                          from the sample app if you want your client app's file tree control to look like Vista's explorer. 
                          3. Removed VC 6 style appwizard comments from the code.
                          4. General review and cleanup of all the code
                          5. The code has now been updated to support VC 2005 or later only. 
                          6. Optimized CFileTreeCtrlItemInfo copy constructor code
                          7. Removed CFileTreeCtrl::SetFlags method as it did not add any functionality to the class.
                          8. DRIVE_ATTRIBUTE_* defines are now enums in the CFileTreeCtrl class.
                          9. Fixed bug in DisplayPath where if the desktop does not contain any files or folders, you would not get a plus
                          button for the desktop. Thanks to Steven Tomer for reporting this bug.
                          10. Prev and Next item stacks now store HTREEITEM's directly in their arrays.
                          11. The control now supports showing the current users folder instead of (or in addition to) the "My Documents" 
                          folder, similar to how the tree control behaves in Windows Explorer on Vista. Note that the control does not 
                          show the "Public" folder which Vista's Windows Explorer shows, because this is implemented as a shell extension 
                          whereas this control is a ground up implementation of a file tree control which does not use the Shell API's
                          for enumerating the Shell Namespace.
                          12. All the classes have been renamed from CTreeFileCtrl... to CFileTreeCtrl...
         PJN / 07-07-2008 1. Code now compiles cleanly using Code Analysis (/analyze)
                          2. Updated code to compile correctly using _ATL_CSTRING_EXPLICIT_CONSTRUCTORS define
                          3. Shares shown underneath the Network node no longer show the share icon overlap. This is consistent with how
                          Windows Explorer shows these items.
                          4. Fixed a bug where CFileTreeCtrl::OnChange was not being called due to the fact that its message map entry
                          was declared incorrectly. Thanks to "Roland" for reporting this bug.
                          5. Fixed a bug in CFileTreeCtrl::SetSelectedPath where the expansion state of a folder was not being correctly
                          set.
                          6. Updated the sample app to allow the expansion state of the selected path to be set.
                          7. Fixed an issue where files and folders were not being displayed when the desktop node was being reexpanded.
                          8. Fixed an issue setting up m_sFQPath for some network node items in EnumNetwork
                          9. During testing of the SetSelectedPath and because the control primarily uses file paths to identify a node,
                          refresh of items which do not have a corresponding path will fail in OnRefresh. Examples of this include 
                          "My Network Places" and "My Computer" amongst others. Until I develop a new version of the control based on 
                          Shell PIDLs, this will be a deficiency of the current approach taken by CFileTreeCtrl
                          10. Following analysis of the auto refresh behaviour of Windows Explorer, the control now only monitors 
                          notifications for changes in the current directory and not all drive letters or recursively from the current
                          directory. This helps to improve the robustness of control when change notifications are received
         PJN / 13-12-2008 1. Optimized the memory usage by removing the m_bFileItem member variable from CFileTreeCtrlItemInfo. Instead
                          now whether an item is a File (as opposed to a folder) is encoded into the m_NodeType value. In addition the
                          m_nodeType value has been made a BYTE value. This helps to save an additional 4 bytes per node. Thanks to 
                          Michael Mtje for prompting this update.
         PJN / 08-01-2009 1. Updated copyright details.
                          2. Fixed a bug in DisplayPath when it calls InsertFileItem to show the root folder. The 
                          CFileTreeCtrlItemInfo pointer passed to InsertFileItem was not being initialized with a valid
                          m_NodeType value. Thanks to Colin Bouckaert for reporting this bug.
                          3. Fixed a bug in SetSelectedPath if you are trying to select a path when SetRootFolder and 
                          SetShowRootedFolder have previously been called. Thanks to Colin Bouckaert for reporting this 
                          bug.
         PJN / 06-04-2013 1. Updated copyright details.
                          2. Updated the code to compile cleanly on VC 2010 and later
                          3. Reworked CFileTreeCtrl::AddMultipleFileNameMasks to use CString::Tokenize
                          4. Removed old Win9x code path from CShareEnumerator class.
                          5. Reworked the CSystemImageList class to remove the defunct Win9x code path.
                          6. Removed the static instances of CSystemImageList and CShareEnumerator. These changes along with
                          the Win9x changes means that the code no longer supports running on Win9x.
                          7. Reworked the code so that calling SetRootFolder on a UNC server name now works. The issue was that a 
                          path of the form "\\MACHINENAME" is not a true file system path because only paths of the form 
                          "\\MACHINENAME\SHARE" allow file enumeration. The code now explicitly checks the path and instead enumerates 
                          the network resources below the UNC server name even when SetRootFolder has been called. Thanks to Jan 
                          Moorman for reporting this issue.
         PJN / 07-04-2013 1. Minor update to the code in DisplayNetworkComputer method.
                          2. Fixed a spelling mistake in EnumNetwork method
         PJN / 06-06-2016 1. Updated copyright details.
                          2. Updated the sample project settings to more modern default values.
                          3. Updated the code to clean compile on VC 2013 - VC 2015.
                          4. Fixed a security issue in the sample app in the use of LoadLibrary on "UXTheme.dll".
                          5. Updated the dialog based sample app's main icon.
                          6. The code now gracefully handles the situation where calls to SHGetDesktopFolder or SHGetMalloc
                          fail during control initialization.
                          7. Fixed a bug where you could not drag an item to a network share.
                          8. The code now shows shortcuts using an overlay mask. This is similar to how shared folders are optionally 
                          already shown by the control with an overlay mask.
                          9. Fixed a issue where the Current User Folder was not showing with a customized icon.
                          10. Fixed an issue in the sample app where the control would fail to show some icons initially due to the
                          fact that the sample app was not initializing OLE on startup.
                          11. Added SAL annotations to all the code
         PJN / 21-07-2019 1. Replaced NULL with nullptr throughout the code.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that the code will now only 
                          compile on VC 2017 or later.
                          3. Replaced enum with enum class throughout the code.
                          4. Replaced CStringArray and CArray with std::vector throughout the code.
                          5. Implemented missing void CFileTreeCtrl::SetEncryptedColor method
                          6. Removed CFileTreeCtrl::IsWindowsVistaOrLater method from the code
                          7. Code now uses ATL::CComPtr to house the COM pointers which it needs
                          8. Code now uses std::atomic<bool> for variables which need to be accessed from multiple threads
                          9. The default color for encrypted files has now changed to be consistent with the color Explorer uses on
                          modern versions of Windows
         PJN / 29-09-2019 1. Fixed a number of compiler warnings when the code is compiled with VS 2019 Preview
         PJN / 21-12-2019 1. Fixed various Clang-Tidy static code analysis warnings in the code.
                          2. Updated initialization of various structs to use C++ 11 list initialization
         PJN / 14-03-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.

Copyright (c) 1999 - 2020 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


/////////////////////////////////  Includes  //////////////////////////////////

#include "stdafx.h"
#ifndef FILETREECTRL_NO_RESOURCES
#include "resource.h"
#endif //#ifndef FILETREECTRL_NO_RESOURCES
#include "FileTreeCtrl.h"


//////////////////////////////// Defines / Locals /////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG

#pragma warning(suppress: 26426)
const UINT WM_TREEUPDATE_CHANGE = ::RegisterWindowMessage(_T("FileTreeCtrlUpdateChange"));
constexpr int FILETREECTRL_TIMERID = 1;
constexpr int FILETREECTRL_TIMERTICK_FOR_EXPANSION = 3;

//Pull in the libs automatically
#pragma comment(lib, "mpr.lib")
#pragma comment(lib, "Netapi32.lib")


////////////////////////////// Implementation /////////////////////////////////

CFileTreeCtrlItemInfo::CFileTreeCtrlItemInfo() noexcept : m_NodeType(NodeType::UnknownNode),
                                                          m_bExtensionHidden(false),
                                                          m_dwNetResourceScope(0),
                                                          m_dwNetResourceType(0),
                                                          m_dwNetResourceDisplayType(0),
                                                          m_dwNetResourceUsage(0)
{
}


#pragma warning(suppress: 26455)
CSystemImageList::CSystemImageList()
{
  //Get the temp directory. This is used to then bring back the system image list
  CString sTempDir;
  GetTempPath(_MAX_PATH, sTempDir.GetBuffer(_MAX_PATH));
  sTempDir.ReleaseBuffer();
  CString sDrive;
  _tsplitpath_s(sTempDir, sDrive.GetBuffer(_MAX_DRIVE), _MAX_DRIVE, nullptr, 0, nullptr, 0, nullptr, 0);
  sDrive.ReleaseBuffer();
  const int nLen = sDrive.GetLength();
  if (nLen)
  {
    if (sDrive[nLen-1] != _T('\\'))
      sDrive += _T("\\");

    //Attach to the system image list
    SHFILEINFO sfi{};
#pragma warning(suppress: 26490)
    Attach(reinterpret_cast<HIMAGELIST>(SHGetFileInfo(sTempDir, 0, &sfi, sizeof(SHFILEINFO), SHGFI_SYSICONINDEX | SHGFI_SMALLICON)));
  }
}


CShareEnumerator::CShareEnumerator() noexcept : m_pNTShareInfo(nullptr),
                                                m_dwShares(0)

{
  //Update the array of shares we know about
  Refresh();
}

CShareEnumerator::~CShareEnumerator()
{
  //Free the buffer if valid
  if (m_pNTShareInfo != nullptr)
  {
    NetApiBufferFree(m_pNTShareInfo);
    m_pNTShareInfo = nullptr;
  }
}

void CShareEnumerator::Refresh() noexcept
{
  //Reset the count of shares
  m_dwShares = 0;

  //Free the buffer if valid
  if (m_pNTShareInfo != nullptr)
    NetApiBufferFree(m_pNTShareInfo);

  //Call the function to enumerate the shares
  DWORD dwEntriesRead = 0;
#pragma warning(suppress: 26490)
  const NET_API_STATUS nStatus = NetShareEnum(nullptr, 502, reinterpret_cast<LPBYTE*>(&m_pNTShareInfo), MAX_PREFERRED_LENGTH, &dwEntriesRead, &m_dwShares, nullptr);
  UNREFERENCED_PARAMETER(nStatus);
}

bool CShareEnumerator::IsShared(_In_ const CString& sPath)
{
  if (m_pNTShareInfo != nullptr)
  {
    for (DWORD i=0; i<m_dwShares; i++)
    {
#pragma warning(suppress: 26481)
      CString sShare(m_pNTShareInfo[i].shi502_path);
#pragma warning(suppress: 26481)
      if ((sPath.CompareNoCase(sShare) == 0) && ((m_pNTShareInfo[i].shi502_type == STYPE_DISKTREE) || ((m_pNTShareInfo[i].shi502_type == STYPE_PRINTQ))))
        return true;
    }
  }
  return false;
}


#pragma warning(suppress: 26409 26440 26433 26477)
IMPLEMENT_DYNCREATE(CFileTreeCtrl, FILETREECTRL_BASE_CLASS)

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CFileTreeCtrl, FILETREECTRL_BASE_CLASS) //NOLINT(modernize-avoid-c-arrays)
#ifndef FILETREECTRL_NO_RESOURCES
  ON_COMMAND(ID_TREEFILECTRL_PROPERTIES, &CFileTreeCtrl::OnProperties)
  ON_UPDATE_COMMAND_UI(ID_TREEFILECTRL_PROPERTIES, &CFileTreeCtrl::OnUpdateProperties)
  ON_COMMAND(ID_TREEFILECTRL_RENAME, &CFileTreeCtrl::OnRename)
  ON_UPDATE_COMMAND_UI(ID_TREEFILECTRL_RENAME, &CFileTreeCtrl::OnUpdateRename)
  ON_COMMAND(ID_TREEFILECTRL_OPEN, &CFileTreeCtrl::OnOpen)
  ON_UPDATE_COMMAND_UI(ID_TREEFILECTRL_OPEN, &CFileTreeCtrl::OnUpdateOpen)
  ON_COMMAND(ID_TREEFILECTRL_DELETE, &CFileTreeCtrl::OnDelete)
  ON_UPDATE_COMMAND_UI(ID_TREEFILECTRL_DELETE, &CFileTreeCtrl::OnUpdateDelete)
  ON_COMMAND(ID_TREEFILECTRL_CREATEFOLDER, &CFileTreeCtrl::OnCreateFolder)
  ON_UPDATE_COMMAND_UI(ID_TREEFILECTRL_CREATEFOLDER, &CFileTreeCtrl::OnUpdateCreateFolder)
  ON_COMMAND(ID_TREEFILECTRL_REFRESH, &CFileTreeCtrl::OnRefresh)
  ON_COMMAND(ID_TREEFILECTRL_UPONELEVEL, &CFileTreeCtrl::OnUpOneLevel)
  ON_UPDATE_COMMAND_UI(ID_TREEFILECTRL_UPONELEVEL, &CFileTreeCtrl::OnUpdateUpOneLevel)
  ON_COMMAND(ID_TREEFILECTRL_BACK, &CFileTreeCtrl::OnBack)
  ON_UPDATE_COMMAND_UI(ID_TREEFILECTRL_BACK, &CFileTreeCtrl::OnUpdateBack)
  ON_COMMAND(ID_TREEFILECTRL_FORWARD, &CFileTreeCtrl::OnForward)
  ON_UPDATE_COMMAND_UI(ID_TREEFILECTRL_FORWARD, &CFileTreeCtrl::OnUpdateForward)
  ON_WM_CONTEXTMENU()
  ON_WM_MOUSEMOVE()
  ON_WM_LBUTTONUP()
  ON_WM_TIMER()
#endif //#ifndef FILETREECTRL_NO_RESOURCES
  ON_WM_INITMENUPOPUP()
  ON_WM_DESTROY()
  ON_WM_CONTEXTMENU()
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT_EX(NM_DBLCLK, &CFileTreeCtrl::OnDblclk)
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT_EX(TVN_ITEMEXPANDING, &CFileTreeCtrl::OnItemExpanding)
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT_EX(TVN_BEGINLABELEDIT, &CFileTreeCtrl::OnBeginLabelEdit)
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT_EX(TVN_ENDLABELEDIT, &CFileTreeCtrl::OnEndLabelEdit)
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT_EX(NM_RCLICK, &CFileTreeCtrl::OnRclick)
#ifndef FILETREECTRL_NO_RESOURCES
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT_EX(TVN_BEGINDRAG, &CFileTreeCtrl::OnBeginDrag)
#endif //#ifndef FILETREECTRL_NO_RESOURCES
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT_EX(NM_CUSTOMDRAW, &CFileTreeCtrl::OnCustomDraw)
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT_EX(TVN_SELCHANGED, &CFileTreeCtrl::OnSelChanged)
#pragma warning(suppress: 26454)
  ON_NOTIFY_REFLECT_EX(TVN_DELETEITEM, &CFileTreeCtrl::OnDeleteItem)
  ON_REGISTERED_MESSAGE(WM_TREEUPDATE_CHANGE, &CFileTreeCtrl::OnChange)
END_MESSAGE_MAP()

#pragma warning(suppress: 26455 26495)
CFileTreeCtrl::CFileTreeCtrl() : FILETREECTRL_BASE_CLASS(),
                                 m_bShowFiles(true),
                                 m_hRoot(TVI_ROOT),
                               #ifndef FILETREECTRL_NO_RESOURCES
                                 m_hItemDrag(nullptr),
                                 m_hItemDrop(nullptr),
                               #endif //#ifndef FILETREECTRL_NO_RESOURCES
                                 m_hNetworkRoot(nullptr),
                                 m_hDesktop(nullptr),
                                 m_hMyDocumentsRoot(nullptr),
                                 m_hMyComputerRoot(nullptr),
                                 m_hCurrentUserFolderRoot(nullptr),
                                 m_hRootedFolder(nullptr),
                                 m_bShowMyComputer(true),
                                 m_bShowDesktop(true),
                                 m_bShowMyDocuments(false),
                                 m_bShowCurrentUserFolder(true),
                               #ifndef FILETREECTRL_NO_RESOURCES
                                 m_pilDrag(nullptr),
                               #endif //#ifndef FILETREECTRL_NO_RESOURCES
                                 m_nTimerID(0),
                               #ifndef FILETREECTRL_NO_RESOURCES
                                 m_DropCopyCursor(nullptr),
                                 m_NoDropCopyCursor(nullptr),
                                 m_DropMoveCursor(nullptr),
                                 m_NoDropMoveCursor(nullptr),
                               #endif //#ifndef FILETREECTRL_NO_RESOURCES
                                 m_TimerTicks(0),
                                 m_bAllowDragDrop(false),
                                 m_bAllowRename(false),
                                 m_bAllowOpen(true),
                                 m_bAllowProperties(true),
                                 m_bAllowDelete(false),
                                 m_bAllowCreateFolder(false),
                                 m_dwDriveHideFlags(0),
                                 m_dwFileHideFlags(FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM | FILE_ATTRIBUTE_OFFLINE | FILE_ATTRIBUTE_TEMPORARY),
                                 m_dwFolderHideFlags(FILE_ATTRIBUTE_HIDDEN | FILE_ATTRIBUTE_SYSTEM | FILE_ATTRIBUTE_OFFLINE | FILE_ATTRIBUTE_TEMPORARY),
                                 m_rgbCompressed(RGB(0, 0, 255)),
                                 m_bShowCompressedUsingDifferentColor(true),
                                 m_rgbEncrypted(RGB(98, 146, 13)),
                                 m_bShowEncryptedUsingDifferentColor(true),
                                 m_nMaxHistory(20),
                                 m_bUpdatingHistorySelection(false),
                                 m_bShowNetwork(true),
                                 m_bShowSharedUsingDifferentIcon(true),
                                 m_FileExtensions(HideFileExtension::UseTheShellSetting),
                                 m_dwNetworkItemTypes(RESOURCETYPE_ANY),
                                 m_bShowDriveLabels(true),
                                 m_bShowRootedFolder(false),
                                 m_bAutoRefresh(false),
                               #ifndef FILETREECTRL_NO_CACHE_COMPUTER_ICON
                                 m_nComputerIconIndex(-1),
                                 m_nSelectedComputerIconIndex(-1),
                               #endif
                                 m_pMonitoringThread(nullptr)
{
  m_FileNameMasks.emplace_back(_T("*.*"));

  for (auto& media : m_MediaIDs)
    media = 0xFFFFFFFF;

  //As a performance optimization we cache the indexes for the computer and selected computer icons, so that we can use it 
  //when displaying machines on the network. This helps to speed up the performance of CFileTreeCtrl when browsing domains or 
  //workgroups with a large number of workstations and servers.
#ifndef FILETREECTRL_NO_CACHE_COMPUTER_ICON
  CString sComputerName;
  DWORD dwCompuerNameSize = MAX_COMPUTERNAME_LENGTH + 1;
  const BOOL bGetCN = GetComputerName(sComputerName.GetBuffer(MAX_COMPUTERNAME_LENGTH + 1), &dwCompuerNameSize);
  sComputerName.ReleaseBuffer();
  if (bGetCN)
  {
    sComputerName = _T("\\\\") + sComputerName;
    m_nComputerIconIndex = GetIconIndex(sComputerName);
    m_nSelectedComputerIconIndex = GetSelIconIndex(sComputerName);
  }
#endif //#ifndef FILETREECTRL_NO_CACHE_COMPUTER_ICON  
}

int CALLBACK CFileTreeCtrl::CompareByFilenameNoCase(_In_ LPARAM lParam1, _In_ LPARAM lParam2, _In_ LPARAM /*lParamSort*/)
{
#pragma warning(suppress: 26490)
  const CFileTreeCtrlItemInfo* pItem1 = reinterpret_cast<CFileTreeCtrlItemInfo*>(lParam1);
#pragma warning(suppress: 26490)
  const CFileTreeCtrlItemInfo* pItem2 = reinterpret_cast<CFileTreeCtrlItemInfo*>(lParam2);
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(pItem1 != nullptr);
#pragma warning(suppress: 26496)
  AFXASSUME(pItem2 != nullptr);

  //Handle the special ordering of some nodes
#pragma warning(suppress: 26489)
  if (pItem1->m_NodeType == CFileTreeCtrlItemInfo::NodeType::CurrentUserFolderNode) //Current User node come before anything else
    return -1;
#pragma warning(suppress: 26489)
  if ((pItem1->m_NodeType == CFileTreeCtrlItemInfo::NodeType::MyDocumentsNode) && (pItem2->m_NodeType != CFileTreeCtrlItemInfo::NodeType::CurrentUserFolderNode)) //Then "My Documents"
    return -1;
#pragma warning(suppress: 26489)
  else if ((pItem1->m_NodeType == CFileTreeCtrlItemInfo::NodeType::MyComputerNode) && (pItem2->m_NodeType != CFileTreeCtrlItemInfo::NodeType::CurrentUserFolderNode) && (pItem2->m_NodeType != CFileTreeCtrlItemInfo::NodeType::MyDocumentsNode)) //Then "My Computer/This PC"
    return -1;
#pragma warning(suppress: 26489)
  else if ((pItem1->m_NodeType == CFileTreeCtrlItemInfo::NodeType::NetworkNode) && (pItem2->m_NodeType != CFileTreeCtrlItemInfo::NodeType::CurrentUserFolderNode) && (pItem2->m_NodeType != CFileTreeCtrlItemInfo::NodeType::MyDocumentsNode) && (pItem2->m_NodeType != CFileTreeCtrlItemInfo::NodeType::MyComputerNode)) //Then "My Network Places" 
    return -1;

  //And then fall back to the standard sorting order
#pragma warning(suppress: 26489)
  if (pItem1->m_NodeType == CFileTreeCtrlItemInfo::NodeType::FileNode)
  {
#pragma warning(suppress: 26489)
    if (pItem2->m_NodeType == CFileTreeCtrlItemInfo::NodeType::FileNode)
      return pItem1->m_sRelativePath.CompareNoCase(pItem2->m_sRelativePath);
    else
      return 1;
  }
  else
  {
#pragma warning(suppress: 26489)
    if (pItem2->m_NodeType == CFileTreeCtrlItemInfo::NodeType::FileNode)
      return -1;
    else
#pragma warning(suppress: 26489)
      return pItem1->m_sRelativePath.CompareNoCase(pItem2->m_sRelativePath);
  }
}

#ifdef _DEBUG
void CFileTreeCtrl::AssertValid() const
{
  //Let the base class do its thing
  __super::AssertValid();
}

void CFileTreeCtrl::Dump(CDumpContext& dc) const
{
  //Let the base class do its thing
  __super::Dump(dc);
}
#endif //#ifdef _DEBUG

void CFileTreeCtrl::SetShowFiles(_In_ bool bFiles)
{
  //Since it can be touched by more than one thead
  m_bShowFiles = bFiles;

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::SetRootFolder(_In_ const CString& sPath)
{
  m_sRootFolder = sPath;

  //Ensure it is not terminated with a "\"
  const int nLength = m_sRootFolder.GetLength();
  if (nLength && (m_sRootFolder[nLength - 1] == _T('\\'))) //NOLINT(clang-analyzer-core.CallAndMessage)
      m_sRootFolder = m_sRootFolder.Left(nLength - 1);

  //Force a refresh
  Refresh();
}

int CFileTreeCtrl::GetIconIndex(_In_ HTREEITEM hItem)
{
  TVITEM tvi{};
  tvi.mask = TVIF_HANDLE | TVIF_IMAGE;
  tvi.hItem = hItem;
  if (GetItem(&tvi))
    return tvi.iImage;
  else
    return -1;
}

#pragma warning(suppress: 26440)
int CFileTreeCtrl::GetIconIndex(_In_ const CString& sFilename)
{
  SHFILEINFO sfi{};
  SHGetFileInfo(sFilename, 0, &sfi, sizeof(SHFILEINFO), SHGFI_SYSICONINDEX | SHGFI_SMALLICON);
  return sfi.iIcon;
}

#pragma warning(suppress: 26440)
int CFileTreeCtrl::GetSelIconIndex(_In_ const CString& sFilename)
{
  SHFILEINFO sfi{};
  SHGetFileInfo(sFilename, 0, &sfi, sizeof(SHFILEINFO), SHGFI_SYSICONINDEX | SHGFI_OPENICON | SHGFI_SMALLICON);
  return sfi.iIcon;
}

#pragma warning(suppress: 26440)
int CFileTreeCtrl::GetIconIndex(_In_ LPITEMIDLIST lpPIDL)
{
  SHFILEINFO sfi{};
#pragma warning(suppress: 26490)
  SHGetFileInfo(reinterpret_cast<LPCTSTR>(lpPIDL), 0, &sfi, sizeof(sfi), SHGFI_PIDL | SHGFI_SYSICONINDEX | SHGFI_SMALLICON);
  return sfi.iIcon;
}

#pragma warning(suppress: 26440)
int CFileTreeCtrl::GetSelIconIndex(_In_ LPITEMIDLIST lpPIDL)
{
  SHFILEINFO sfi{};
#pragma warning(suppress: 26490)
  SHGetFileInfo(reinterpret_cast<LPCTSTR>(lpPIDL), 0, &sfi, sizeof(sfi), SHGFI_PIDL | SHGFI_SYSICONINDEX | SHGFI_SMALLICON | SHGFI_OPENICON);
  return sfi.iIcon;
}

int CFileTreeCtrl::GetSelIconIndex(_In_ HTREEITEM hItem)
{
  TVITEM tvi{};
  tvi.mask = TVIF_HANDLE | TVIF_SELECTEDIMAGE;
  tvi.hItem = hItem;
  if (GetItem(&tvi))
    return tvi.iSelectedImage;
  else
    return -1;
}

HTREEITEM CFileTreeCtrl::FindSibling(_In_ HTREEITEM hParent, _In_ const CString& sItem) const
{
  HTREEITEM hChild = GetChildItem(hParent);
  while (hChild)
  {
#pragma warning(suppress: 26490)
    const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(hChild));
#pragma warning(suppress: 26496)
    AFXASSUME(pItem != nullptr);
#pragma warning(suppress: 26489)
    if (pItem->m_sRelativePath.CompareNoCase(sItem) == 0)
      return hChild;
    hChild = GetNextItem(hChild, TVGN_NEXT);
  }
  return nullptr;
}

CString CFileTreeCtrl::GetSelectedPath()
{
  HTREEITEM hItem = GetSelectedItem();
  if (hItem)
    return ItemToPath(hItem);
  else
    return CString();
}

HTREEITEM CFileTreeCtrl::FindServersNode(_In_ HTREEITEM hFindFrom) const
{
  if (m_bShowNetwork)
  {
    //Try to find some "servers" in the child items of hFindFrom
    HTREEITEM hChild = GetChildItem(hFindFrom);
    while (hChild)
    {
#pragma warning(suppress: 26490)
      const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(hChild));
#pragma warning(suppress: 26496)
      AFXASSUME(pItem != nullptr);
      //Found a share
#pragma warning(suppress: 26489)
      if (pItem->m_dwNetResourceDisplayType == RESOURCEDISPLAYTYPE_SERVER)
        return hFindFrom;

      //Get the next sibling for the next loop around
      hChild = GetNextSiblingItem(hChild);
    }

    //Ok, since we got here, we did not find any servers in any of the child nodes of this 
    //item. In this case we need to call ourselves recursively to find one
    hChild = GetChildItem(hFindFrom);
    while (hChild)
    {
      HTREEITEM hFound = FindServersNode(hChild);
      if (hFound)
        return hFound;

      //Get the next sibling for the next loop around
      hChild = GetNextSiblingItem(hChild);
    }
  }

  //If we got as far as here then no servers were found.
  return nullptr;
}

void CFileTreeCtrl::SetHasPlusButton(_In_ HTREEITEM hItem, _In_ bool bHavePlus)
{
  TVITEM tvItem{};
  tvItem.hItem = hItem;
  tvItem.mask = TVIF_HANDLE | TVIF_CHILDREN;
  tvItem.cChildren = bHavePlus;
  SetItem(&tvItem);
}

bool CFileTreeCtrl::HasPlusButton(_In_ HTREEITEM hItem)
{
  TVITEM tvItem{};
  tvItem.hItem = hItem;
  tvItem.mask = TVIF_HANDLE | TVIF_CHILDREN;
  return GetItem(&tvItem) && (tvItem.cChildren != 0);
}

HTREEITEM CFileTreeCtrl::SetSelectedPath(_In_ const CString& sPath, _In_ bool bExpanded)
{
  CString sSearch(sPath);
  sSearch.MakeUpper();
  int nSearchLength = sSearch.GetLength();
  if (nSearchLength == 0)
  {
    TRACE(_T("CFileTreeCtrl::SetSelectedPath, Cannot select a empty path\n"));
    return nullptr;
  }

  //Remove initial part of path if the root folder is setup
  CString sRootFolder(m_sRootFolder);
  sRootFolder.MakeUpper();
  const int nRootLength = sRootFolder.GetLength();
  if (nRootLength)
  {
    if (sSearch.Find(sRootFolder) != 0)
    {
      TRACE(_T("CFileTreeCtrl::SetSelectedPath, Could not select the path %s as the root has been configued as %s\n"), sPath.GetString(), m_sRootFolder.GetString());
      return nullptr;
    }
    sSearch = sSearch.Right(sSearch.GetLength() - nRootLength);

    //Remove leading "\" from the path
    nSearchLength = sSearch.GetLength();
    if (nSearchLength > 1 && sSearch.GetAt(0) == _T('\\'))
      sSearch = sSearch.Right(nSearchLength-1);
  }

  //Remove trailing "\" from the path
  nSearchLength = sSearch.GetLength();
  if ((nSearchLength > 3) && sSearch.GetAt(nSearchLength - 1) == _T('\\'))
    sSearch = sSearch.Left(nSearchLength-1);

  SetRedraw(FALSE);

  //Handle selection of the rooted folder if necessary
  HTREEITEM hItemFound = m_hRoot;
  if (nRootLength && m_hRootedFolder)
  {
    hItemFound = m_hRootedFolder;
    SelectItem(hItemFound);

    const bool bHasPlus = HasPlusButton(hItemFound);
    const bool bHasChildren = (GetChildItem(hItemFound) != nullptr);

    if (bHasPlus && !bHasChildren)
      DoExpand(hItemFound);
    else
      Expand(hItemFound, TVE_EXPAND);

    if (sSearch.IsEmpty())
    {
      SetRedraw(TRUE);
      return hItemFound;
    }
  }

  if (sSearch.IsEmpty())
  {
    SetRedraw(TRUE);
    return nullptr;
  }

  bool bDriveMatch = sRootFolder.IsEmpty();
  const bool bNetworkMatch = m_bShowNetwork && ((sSearch.GetLength() > 2) && sSearch.Find(_T("\\\\")) == 0);
  if (bNetworkMatch)
  {
    bDriveMatch = false;

    const bool bHasPlus = HasPlusButton(m_hNetworkRoot);
    const bool bHasChildren = (GetChildItem(m_hNetworkRoot) != nullptr);

    if (bHasPlus && !bHasChildren)
      DoExpand(m_hNetworkRoot);
    else
      Expand(m_hNetworkRoot, TVE_EXPAND);
    hItemFound = FindServersNode(m_hNetworkRoot);
    sSearch = sSearch.Right(sSearch.GetLength() - 2);
  }
  if (bDriveMatch)
  {
    CString sMyDocumentsPath(m_sMyDocumentsPath);
    sMyDocumentsPath.MakeUpper();
    CString sMyDesktopPath(m_sDesktopPath);
    sMyDesktopPath.MakeUpper();
    CString sCurrentUserFolderPath(m_sCurrentUserFolderPath);
    sCurrentUserFolderPath.MakeUpper();

    if (m_hDesktop && sSearch.Find(sMyDesktopPath) == 0)
    {
      const bool bHasPlus = HasPlusButton(m_hDesktop);
      const bool bHasChildren = (GetChildItem(m_hDesktop) != nullptr);

      if (bHasPlus && !bHasChildren)
        DoExpand(m_hDesktop);
      else
        Expand(m_hDesktop, TVE_EXPAND);
      hItemFound = m_hDesktop;

      //Skip over the part of the search path which contains the desktop path
      sSearch = sSearch.Right(sSearch.GetLength() - sMyDesktopPath.GetLength() - 1);

      //Because of the special way that the desktop item is shown, check to see if
      //we should collapse the desktop node
      if (sSearch.GetLength() == 0)
        Expand(m_hDesktop, TVE_COLLAPSE);

      bDriveMatch = false;
    }
    else if (m_hCurrentUserFolderRoot && sSearch.Find(sCurrentUserFolderPath) == 0)
    {
      const bool bHasPlus = HasPlusButton(m_hCurrentUserFolderRoot);
      const bool bHasChildren = (GetChildItem(m_hCurrentUserFolderRoot) != nullptr);

      if (bHasPlus && !bHasChildren)
        DoExpand(m_hCurrentUserFolderRoot);
      else
        Expand(m_hCurrentUserFolderRoot, TVE_EXPAND);
      hItemFound = m_hCurrentUserFolderRoot;

      //Skip over the part of the search path which contains the Current Users path
      sSearch = sSearch.Right(sSearch.GetLength() - sCurrentUserFolderPath.GetLength() - 1);
      bDriveMatch = false;
    }
    else if (m_hMyDocumentsRoot && sSearch.Find(sMyDocumentsPath) == 0)
    {
      const bool bHasPlus = HasPlusButton(m_hMyDocumentsRoot);
      const bool bHasChildren = (GetChildItem(m_hMyDocumentsRoot) != nullptr);

      if (bHasPlus && !bHasChildren)
        DoExpand(m_hMyDocumentsRoot);
      else
        Expand(m_hMyDocumentsRoot, TVE_EXPAND);
      hItemFound = m_hMyDocumentsRoot;

      //Skip over the part of the search path which contains the My documents path
      sSearch = sSearch.Right(sSearch.GetLength() - sMyDocumentsPath.GetLength() - 1);
      bDriveMatch = false;
    }
    else if (m_hMyComputerRoot)
    {
      const bool bHasPlus = HasPlusButton(m_hMyComputerRoot);
      const bool bHasChildren = (GetChildItem(m_hMyComputerRoot) != nullptr);

      if (bHasPlus && !bHasChildren)
        DoExpand(m_hMyComputerRoot);
      else
        Expand(m_hMyComputerRoot, TVE_EXPAND);
      hItemFound = m_hMyComputerRoot;
    }
  }

  int nFound = sSearch.Find(_T('\\'));
  while (nFound != -1)
  {
    CString sMatch;
    if (bDriveMatch)
    {
      sMatch = sSearch.Left(nFound + 1);
      bDriveMatch = false;
    }
    else
      sMatch = sSearch.Left(nFound);

    hItemFound = FindSibling(hItemFound, sMatch);
    if (hItemFound == nullptr)
      break;
    else if (!IsDrive(sPath))
    {
      ASSERT(hItemFound != TVI_ROOT);
      SelectItem(hItemFound);

      const bool bHasPlus = HasPlusButton(hItemFound);
      const bool bHasChildren = (GetChildItem(hItemFound) != nullptr);

      if (bHasPlus && !bHasChildren)
        DoExpand(hItemFound);
      else
        Expand(hItemFound, TVE_EXPAND);
    }

    sSearch = sSearch.Right(sSearch.GetLength() - nFound - 1);
    nFound = sSearch.Find(_T('\\'));
  };

  //The last item
  if (hItemFound)
  {
    if (sSearch.GetLength())
    {
      HTREEITEM hTempItem = FindSibling(hItemFound, sSearch);
      if (hTempItem)
        hItemFound = hTempItem;
      else
      {
        //If we cannot find the final item, then deliberately do not expand the parent folder
        bExpanded = false;
      }
    }
    ASSERT(hItemFound != TVI_ROOT);
    SelectItem(hItemFound);
    if (bExpanded)
    {
      const bool bHasPlus = HasPlusButton(hItemFound);
      const bool bHasChildren = (GetChildItem(hItemFound) != nullptr);

      if (bHasPlus && !bHasChildren)
        DoExpand(hItemFound);
      Expand(hItemFound, TVE_EXPAND);
    }
  }

  //Turn back on the redraw flag
  SetRedraw(TRUE);

  return hItemFound;
}

bool CFileTreeCtrl::Rename(_In_ HTREEITEM hItem)
{
  if (hItem)
  {
#pragma warning(suppress: 26490)
    const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(hItem));
#pragma warning(suppress: 26496)
    AFXASSUME(pItem != nullptr);
#pragma warning(suppress: 26489)
    if (m_bAllowRename && ((pItem->m_NodeType == CFileTreeCtrlItemInfo::NodeType::FileNode) || (pItem->m_NodeType == CFileTreeCtrlItemInfo::NodeType::FolderNode)))
      return (EditLabel(hItem) != nullptr);
    else
      return false;
  }
  else
    return false;
}

bool CFileTreeCtrl::ShowProperties(_In_ HTREEITEM hItem)
{
  bool bSuccess = false;
  if (m_bAllowProperties && hItem)
  {
    //Show the "properties" for the selected file
    CString sFile(ItemToPath(hItem));
#pragma warning(suppress: 26476)
    SHELLEXECUTEINFO sei{};
    sei.cbSize = sizeof(sei);
    sei.hwnd = GetParent()->GetSafeHwnd();
    sei.nShow = SW_SHOW;
    sei.lpFile = sFile.GetBuffer(sFile.GetLength());
    sei.lpVerb = _T("properties");
    sei.fMask  = SEE_MASK_INVOKEIDLIST;
#pragma warning(suppress: 26486)
    bSuccess = ShellExecuteEx(&sei);
    sFile.ReleaseBuffer();
  }
  return bSuccess;
}

CString CFileTreeCtrl::CreateNewChildFolderName(_In_ const CString& sParentFolder)
{
  //What will be the return value from this function
  CString sNewFolder;

  //Remove trailing "\" from the parent folder path
  CString sLocalParentFolder(sParentFolder);
  const int nFind = sLocalParentFolder.GetLength();
  if ((nFind != -1) && sLocalParentFolder.GetAt(nFind - 1) == _T('\\'))
    sLocalParentFolder = sLocalParentFolder.Left(nFind - 1);

  bool bContinue = true;
  int nAttempt = 1;
  while (bContinue)
  {
    //Create the new potential folder name
    if (nAttempt == 1)
      sNewFolder.Format(_T("%s\\New Folder"), sLocalParentFolder.GetString());
    else
      sNewFolder.Format(_T("%s\\New Folder (%d)"), sLocalParentFolder.GetString(), nAttempt);

    //Check to see if it already exists
    const DWORD dwAttributes = GetFileAttributes(sNewFolder);
    if (dwAttributes == INVALID_FILE_ATTRIBUTES)
      bContinue = false;
    else
      ++nAttempt;
  }

  return sNewFolder;
}

HTREEITEM CFileTreeCtrl::CreateFolder(_In_ HTREEITEM hItem, _In_ const CString& sNewFolderName)
{
  //What will be the return value from this function assume the worst
  HTREEITEM hNewFolder = nullptr;

  if (hItem)
  {
#pragma warning(suppress: 26490)
    const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(hItem));
#pragma warning(suppress: 26496)
    AFXASSUME(pItem != nullptr);
#pragma warning(suppress: 26489)
    if (m_bAllowCreateFolder && (pItem->m_NodeType != CFileTreeCtrlItemInfo::NodeType::NetworkNode))
    {
      //Get the path of the parent folder
      CString sParentFolder(ItemToPath(hItem));

      //Form the name of the new folder
      CString sNewFolder(sNewFolderName.IsEmpty() ? CreateNewChildFolderName(sParentFolder) : sNewFolderName);

      //Kill the monitoring thread if necessary
      const bool bMonitoringThread = (m_pMonitoringThread != nullptr);
      CString sMonitoringPath(m_sMonitoringPath);
      if (bMonitoringThread)
        KillMonitoringThread();

      //Try to create the new folder
      if (CreateDirectory(sNewFolder, nullptr))
      {
        //Notify the shell that a new folder has been created
        SHChangeNotify(SHCNE_MKDIR, SHCNF_PATH, sNewFolder, nullptr);
      
        //Expand the parent folder
        Expand(hItem, TVE_EXPAND);

        //Break apart the new folder name ready for setting up the CFileTreeCtrlItemInfo* item
        CString sFname;
        CString sExt;
        _tsplitpath_s(sNewFolder, nullptr, 0, nullptr, 0, sFname.GetBuffer(_MAX_FNAME), _MAX_FNAME, sExt.GetBuffer(_MAX_EXT), _MAX_EXT);
        sExt.ReleaseBuffer();
        sFname.ReleaseBuffer();
        CString sPath;
        _tmakepath_s(sPath.GetBuffer(_MAX_PATH), _MAX_PATH, nullptr, nullptr, sFname, sExt);
        sPath.ReleaseBuffer();

        //Determine if the parent item already has child items before we add the new item
        const bool bHasChildren = (GetChildItem(hItem) != nullptr);

        //Insert the new item
#pragma warning(suppress: 26403 26409)
        auto pNewItem = new CFileTreeCtrlItemInfo;
        pNewItem->m_sFQPath = sNewFolder;
        pNewItem->m_sRelativePath = sPath;
        pNewItem->m_NodeType = CFileTreeCtrlItemInfo::NodeType::FolderNode;
#pragma warning(suppress: 26486)
        hNewFolder = InsertFileItem(hItem, pNewItem, 0, GetIconIndex(sNewFolder), GetSelIconIndex(sNewFolder), true);
        if (hNewFolder)
        {
          //Update the plus button for the parent if necessary
          if (!bHasChildren)
            SetHasPlusButton(hItem, true);

          //Select the new folder
          ASSERT(hNewFolder != TVI_ROOT);
          SelectItem(hNewFolder);

          //Attempt in place editing of the new folder
          Rename(hNewFolder);
        }
      }

      //Recreate the monitoring thread if necessary
      if (bMonitoringThread)
        CreateMonitoringThread(m_sMonitoringPath);
    }
  }

  return hNewFolder;
}

bool CFileTreeCtrl::Delete(_In_ HTREEITEM hItem)
{
  bool bSuccess = false;

  if (hItem)
  {
#pragma warning(suppress: 26490)
    const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(hItem));
#pragma warning(suppress: 26496)
    AFXASSUME(pItem != nullptr);
#pragma warning(suppress: 26489)
    if (m_bAllowDelete && ((pItem->m_NodeType == CFileTreeCtrlItemInfo::NodeType::FileNode) || ((pItem->m_NodeType == CFileTreeCtrlItemInfo::NodeType::FolderNode))))
    {
      SHFILEOPSTRUCT shfo{};
      shfo.hwnd = GetParent()->GetSafeHwnd();
      shfo.wFunc = FO_DELETE;

      //Undo is not allowed if the SHIFT key is held down
      if (!(GetKeyState(VK_SHIFT) & 0x8000))
        shfo.fFlags = FOF_ALLOWUNDO;

      //Create a Multi SZ string with the filename to delete
      CString sFileToDelete(ItemToPath(hItem));
#pragma warning(suppress: 26472)
      size_t nChars = static_cast<size_t>(sFileToDelete.GetLength()) + 1;
      nChars++;
      std::vector<TCHAR> sTempFrom;
      sTempFrom.resize(nChars);
      _tcscpy_s(sTempFrom.data(), nChars, sFileToDelete);
#pragma warning(suppress: 26446)
      sTempFrom[nChars - 1] = _T('\0');
      shfo.pFrom = sTempFrom.data();

      //Kill the monitoring thread if necessary
      const bool bMonitoringThread = (m_pMonitoringThread != nullptr);
      CString sMonitoringPath(m_sMonitoringPath);
      if (bMonitoringThread)
        KillMonitoringThread();

      //Let the shell perform the actual deletion
#pragma warning(suppress: 26486)
      if (SHFileOperation(&shfo) == 0 && (shfo.fAnyOperationsAborted == FALSE))
      {
        //Gather the parent, next sibling and previous sibling before we do the delete
        HTREEITEM hParent = GetParentItem(hItem);
        HTREEITEM hPrevious = GetNextItem(hItem, TVGN_PREVIOUS);
        HTREEITEM hNext = GetNextItem(hItem, TVGN_NEXT);

        //Work out which item we should reselect after the delete
        if (hPrevious == nullptr)
          hPrevious = hNext;
        if (hPrevious == nullptr)
          hPrevious = hParent;

        //Delete the item from the view
        bSuccess = DeleteItem(hItem);

        //Update the selected item and plus button state of the parent if required
        if (bSuccess && hPrevious)
        {
          Select(hPrevious, TVGN_CARET);
          if (hPrevious == hParent)
            SetHasPlusButton(hParent, false);
        }
      }

      //Recreate the monitoring thread if necessary
      if (bMonitoringThread)
        CreateMonitoringThread(m_sMonitoringPath);
    }
  }
  return bSuccess;
}

bool CFileTreeCtrl::Open(_In_ HTREEITEM hItem)
{
  bool bSuccess = false;
  if (m_bAllowOpen && hItem)
  {
    //Execute the default command for the selected item
    CString sFile(ItemToPath(hItem));
#pragma warning(suppress: 26476)
    SHELLEXECUTEINFO sei{};
    sei.cbSize = sizeof(sei);
    sei.hwnd = GetParent()->GetSafeHwnd();
    sei.nShow = SW_SHOW;
    sei.lpFile = sFile.GetBuffer(sFile.GetLength());
    sei.fMask  = SEE_MASK_INVOKEIDLIST;
#pragma warning(suppress: 26486)
    bSuccess = ShellExecuteEx(&sei);
    sFile.ReleaseBuffer();
  }
  return bSuccess;
}

void CFileTreeCtrl::SetDriveHideFlags(_In_ DWORD dwDriveHideFlags)
{
  m_dwDriveHideFlags = dwDriveHideFlags;

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::SetFileHideFlags(_In_ DWORD dwFileHideFlags)
{
  m_dwFileHideFlags = dwFileHideFlags;

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::SetFolderHideFlags(_In_ DWORD dwFolderHideFlags)
{
  m_dwFolderHideFlags = dwFolderHideFlags;

  //Force a refresh
  Refresh();
}

size_t CFileTreeCtrl::AddMultipleFileNameMasks(_In_ const CString& sFileNameMasks)
{
  //Empty out the array
  m_FileNameMasks.clear();

  int nCurPos = 0;
  CString sToken(sFileNameMasks.Tokenize(_T(",;"), nCurPos));
  while (sToken != _T(""))
  {
    sToken.Trim();
    m_FileNameMasks.push_back(sToken);
    sToken = sFileNameMasks.Tokenize(_T(",;"), nCurPos);
  }

  //Force a refresh
  Refresh();

  //Return the number of recipients parsed
  return m_FileNameMasks.size();
}

CString CFileTreeCtrl::GetFileNameMasks() const
{
  //What will be the return value from this function
  CString sMasks;

  const auto nSize = m_FileNameMasks.size();
  for (size_t i=0; i<nSize; i++)
  {
#pragma warning(suppress: 26446)
    sMasks += m_FileNameMasks[i];
    if (i < (nSize-1))
      sMasks += _T(",");
  }
  return sMasks;
}

void CFileTreeCtrl::SetFileNameMask(_In_ size_t nIndex, _In_ const CString& sFileNameMask)
{
#pragma warning(suppress: 26446)
  m_FileNameMasks[nIndex] = sFileNameMask;

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::AddFileNameMask(_In_ const CString& sFileNameMask)
{
  m_FileNameMasks.push_back(sFileNameMask);

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::SetCompressedColor(_In_ COLORREF rgbCompressed)
{
  m_rgbCompressed = rgbCompressed;

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::SetEncryptedColor(_In_ COLORREF rgbEncrypted)
{
  m_rgbEncrypted = rgbEncrypted;

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::SetShowNetwork(_In_ bool bShowNetwork)
{
  m_bShowNetwork = bShowNetwork;

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::SetUsingDifferentColorForCompressed(_In_ bool bShowCompressedUsingDifferentColor)
{
  m_bShowCompressedUsingDifferentColor = bShowCompressedUsingDifferentColor;

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::SetUsingDifferentIconForSharedFolders(_In_ bool bShowSharedUsingDifferentIcon)
{
  m_bShowSharedUsingDifferentIcon = bShowSharedUsingDifferentIcon;

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::SetUsingDifferentColorForEncrypted(_In_ bool bShowEncryptedUsingDifferentColor)
{
  m_bShowEncryptedUsingDifferentColor = bShowEncryptedUsingDifferentColor;

  //Force a refresh
  Refresh();
};

void CFileTreeCtrl::SetShowFileExtensions(_In_ HideFileExtension FileExtensions)
{
  m_FileExtensions = FileExtensions;

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::SetNetworkItemTypes(_In_ DWORD dwTypes)
{
  m_dwNetworkItemTypes = dwTypes;

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::SetShowDriveLabels(_In_ bool bShowDriveLabels)
{
  m_bShowDriveLabels = bShowDriveLabels;

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::SetShowMyComputer(_In_ bool bShowMyComputer)
{
  m_bShowMyComputer = bShowMyComputer;

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::SetShowDesktop(_In_ bool bShowDesktop)
{
  m_bShowDesktop = bShowDesktop;

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::SetShowMyDocuments(_In_ bool bShowMyDocuments)
{
  m_bShowMyDocuments = bShowMyDocuments;

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::SetShowCurrentUserFolder(_In_ bool bShowCurrentUserFolder)
{
  m_bShowCurrentUserFolder = bShowCurrentUserFolder;

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::SetShowRootedFolder(_In_ bool bShowRootedFolder)
{
  m_bShowRootedFolder = bShowRootedFolder;

  //Force a refresh
  Refresh();
}

bool CFileTreeCtrl::CanDisplayFile(_In_ const CFileFind& find)
{
  //Derived classes can decide dynamically whether or not a certain file is to be displayed. CFileTreeCtrl by default
  //displays all files which do not have attributes as set in m_dwFileHideFlags

  return (m_bShowFiles && !find.IsDirectory() && !find.MatchesMask(m_dwFileHideFlags));
}

bool CFileTreeCtrl::CanDisplayFolder(_In_ const CFileFind& find)
{
  //Derived classes can decide dynamically whether or not a certain folder is to be displayed. CFileTreeCtrl by default
  //displays all folders excluding the ".." and "." entries and which do not have attributes as set in m_dwFolderHideFlags

  return (find.IsDirectory() && !find.IsDots() && !find.MatchesMask(m_dwFolderHideFlags));
}

#pragma warning(suppress: 26440)
bool CFileTreeCtrl::CanDisplayNetworkItem(_In_ CFileTreeCtrlItemInfo* /*pItem*/)
{
  //Derived classes can decide dynamically whether or not a certain network item is to be displayed. 
  //CFileTreeCtrl by default displays all items

  return true;
}

bool CFileTreeCtrl::CanHandleChangeNotifications(_In_ const CString& sPath)
{
  //If no path was given, then fail this function
  if (sPath.IsEmpty())
    return false;

  //We are not interested in files
  if (IsFile(sPath))
    return false;

  //The path must be a folder of some kind
  if (!IsFolder(sPath))
    return false;

  //check if this drive is one of the types which can issue notification changes
  CString sDrive(sPath);
  if (!IsDrive(sDrive))
    sDrive = sPath.Left(3);

  const UINT nDrive = GetDriveType(sDrive);
  return ((nDrive != DRIVE_REMOVABLE) && nDrive != DRIVE_CDROM);
}

#pragma warning(suppress: 26440)
bool CFileTreeCtrl::CanDisplayDrive(_In_ const CString& sDrive)
{
  //Derived classes can decide dynamically whether or not a 
  //certain drive is to be displayed. CFileTreeCtrl by default
  //displays all drives which do not have attributes as set in
  //m_dwDriveHideFlags

  //check if this drive is one of the types to hide
  bool bDisplay = true;
  const UINT nDrive = GetDriveType(sDrive);
  switch (nDrive)
  {
    case DRIVE_REMOVABLE:
    {
      if (m_dwDriveHideFlags & static_cast<DWORD>(DriveType::DRIVE_ATTRIBUTE_REMOVABLE))
        bDisplay = false;
      break;
    }
    case DRIVE_FIXED:
    {
      if (m_dwDriveHideFlags & static_cast<DWORD>(DriveType::DRIVE_ATTRIBUTE_FIXED))
        bDisplay = false;
      break;
    }
    case DRIVE_REMOTE:
    {
      if (m_dwDriveHideFlags & static_cast<DWORD>(DriveType::DRIVE_ATTRIBUTE_REMOTE))
        bDisplay = false;
      break;
    }
    case DRIVE_CDROM:
    {
      if (m_dwDriveHideFlags & static_cast<DWORD>(DriveType::DRIVE_ATTRIBUTE_CDROM))
        bDisplay = false;
      break;
    }
    case DRIVE_RAMDISK:
    {
      if (m_dwDriveHideFlags & static_cast<DWORD>(DriveType::DRIVE_ATTRIBUTE_RAMDISK))
        bDisplay = false;
      break;
    }
    default:
    {
      break;
    }
  }

  return bDisplay;
}

void CFileTreeCtrl::OnRename()
{
  Rename(GetSelectedItem());
}

#pragma warning(suppress: 26429)
void CFileTreeCtrl::OnUpdateRename(CCmdUI* pCmdUI)
{
  HTREEITEM hSelItem = GetSelectedItem();
  if (hSelItem)
  {
#pragma warning(suppress: 26490)
    const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(hSelItem));
#pragma warning(suppress: 26496)
    AFXASSUME(pItem != nullptr);
#pragma warning(suppress: 26486 26489)
    pCmdUI->Enable(m_bAllowRename && ((pItem->m_NodeType == CFileTreeCtrlItemInfo::NodeType::FileNode) || (pItem->m_NodeType == CFileTreeCtrlItemInfo::NodeType::FolderNode)));
  }
  else
#pragma warning(suppress: 26486)
    pCmdUI->Enable(FALSE);
}

void CFileTreeCtrl::OnProperties()
{
  ShowProperties(GetSelectedItem());
}

#pragma warning(suppress: 26429)
void CFileTreeCtrl::OnUpdateProperties(CCmdUI* pCmdUI)
{
  HTREEITEM hSelItem = GetSelectedItem();
  if (hSelItem)
  {
    if (m_bAllowProperties)
    {
#pragma warning(suppress: 26490)
      const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(hSelItem));
#pragma warning(suppress: 26496)
      AFXASSUME(pItem != nullptr);
#pragma warning(suppress: 26489)
      if (pItem->m_NodeType == CFileTreeCtrlItemInfo::NodeType::NetworkNode)
      {
#pragma warning(suppress: 26486 26489)
        pCmdUI->Enable((pItem->m_dwNetResourceDisplayType == RESOURCEDISPLAYTYPE_SERVER) || (pItem->m_dwNetResourceDisplayType == RESOURCEDISPLAYTYPE_SHARE));
      }
      else
#pragma warning(suppress: 26486)
        pCmdUI->Enable(TRUE);
    }
    else
#pragma warning(suppress: 26486)
      pCmdUI->Enable(FALSE);
  }
  else
#pragma warning(suppress: 26486)
    pCmdUI->Enable(FALSE);
}

void CFileTreeCtrl::OnOpen()
{
  Open(GetSelectedItem());
}

#pragma warning(suppress: 26429)
void CFileTreeCtrl::OnUpdateOpen(CCmdUI* pCmdUI)
{
  HTREEITEM hSelItem = GetSelectedItem();
  if (hSelItem)
  {
    if (m_bAllowOpen)
    {
#pragma warning(suppress: 26490)
      const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(hSelItem));
#pragma warning(suppress: 26496)
      AFXASSUME(pItem != nullptr);
#pragma warning(suppress: 26489)
      if (pItem->m_NodeType == CFileTreeCtrlItemInfo::NodeType::NetworkNode)
      {
#pragma warning(suppress: 26486 26489)
        pCmdUI->Enable((pItem->m_dwNetResourceDisplayType == RESOURCEDISPLAYTYPE_SERVER) || (pItem->m_dwNetResourceDisplayType == RESOURCEDISPLAYTYPE_SHARE));
      }
      else
#pragma warning(suppress: 26486)
        pCmdUI->Enable(TRUE);
    }
    else
#pragma warning(suppress: 26486)
      pCmdUI->Enable(FALSE);
  }
  else
#pragma warning(suppress: 26486)
    pCmdUI->Enable(FALSE);
}

void CFileTreeCtrl::OnDelete()
{
  Delete(GetSelectedItem());
}

#pragma warning(suppress: 26429)
void CFileTreeCtrl::OnUpdateDelete(CCmdUI* pCmdUI)
{
  HTREEITEM hSelItem = GetSelectedItem();
  if (hSelItem)
  {
#pragma warning(suppress: 26490)
    const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(hSelItem));
#pragma warning(suppress: 26496)
    AFXASSUME(pItem != nullptr);
#pragma warning(suppress: 26486 26489)
    pCmdUI->Enable(m_bAllowDelete && ((pItem->m_NodeType == CFileTreeCtrlItemInfo::NodeType::FileNode) || ((pItem->m_NodeType == CFileTreeCtrlItemInfo::NodeType::FolderNode))));
  }
  else
#pragma warning(suppress: 26486)
    pCmdUI->Enable(FALSE);
}

void CFileTreeCtrl::OnCreateFolder()
{
  CreateFolder(GetSelectedItem());
}

#pragma warning(suppress: 26429)
void CFileTreeCtrl::OnUpdateCreateFolder(CCmdUI* pCmdUI)
{
  HTREEITEM hSelItem = GetSelectedItem();
  if (hSelItem && m_bAllowCreateFolder)
  {
    //Allow creation of a folder if the selected item is itself a directory/folder
#pragma warning(suppress: 26490)
    const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(hSelItem));
    ASSERT(pItem != nullptr);
#pragma warning(suppress: 26489)
    const DWORD dwAttributes = GetFileAttributes(pItem->m_sFQPath);
    if (dwAttributes != INVALID_FILE_ATTRIBUTES)
#pragma warning(suppress: 26486)
      pCmdUI->Enable((dwAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0);
    else
#pragma warning(suppress: 26486)
      pCmdUI->Enable(FALSE);
  }
  else
#pragma warning(suppress: 26486)
    pCmdUI->Enable(FALSE);
}


#ifndef FILETREECTRL_NO_RESOURCES
#pragma warning(suppress: 26434)
void CFileTreeCtrl::OnContextMenu(CWnd*, CPoint point)
{
  CMenu menu;
  if (menu.LoadMenu(IDR_TREEFILECTRL_POPUP))
  {
    CMenu* pPopup = menu.GetSubMenu(0);
#pragma warning(suppress: 26496)
    AFXASSUME(pPopup != nullptr);
    pPopup->TrackPopupMenu(TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y, this);
  }
}
#endif //#ifndef FILETREECTRL_NO_RESOURCES

void CFileTreeCtrl::UpOneLevel()
{
  OnUpOneLevel();
}

#pragma warning(suppress: 26429)
void CFileTreeCtrl::OnUpdateUpOneLevel(CCmdUI* pCmdUI)
{
  HTREEITEM hItem = GetSelectedItem();
  if (hItem)
#pragma warning(suppress: 26486)
    pCmdUI->Enable(GetParentItem(hItem) != nullptr);
  else
#pragma warning(suppress: 26486)
    pCmdUI->Enable(FALSE);
}

#pragma warning(suppress: 26429)
BOOL CFileTreeCtrl::OnDblclk(NMHDR* /*pNMHDR*/, LRESULT* pResult) 
{
  HTREEITEM hItem = GetSelectedItem();
  CPoint pt(GetCurrentMessage()->pt);
  ScreenToClient(&pt);
  if (hItem && (hItem == HitTest(pt)))
  {
    if (!HasPlusButton(hItem))
      OnOpen();
  }
  *pResult = 0;
  return FALSE; //Allow the message to be reflected again
}

bool CFileTreeCtrl::GoBack()
{
  bool bSuccess = false;
  if (m_PrevItems.size())
  {
    OnBack();
    bSuccess = true;
  }
  return bSuccess;
}

bool CFileTreeCtrl::GoForward()
{
  bool bSuccess = false;
  if (m_NextItems.size())
  {
    OnForward();
    bSuccess = true;
  }
  return bSuccess;
}

void CFileTreeCtrl::Refresh()
{
  if (GetSafeHwnd())
    OnRefresh();
}

void CFileTreeCtrl::OnRefresh()
{
  //Just in case this will take some time
  CWaitCursor wait;

  SetRedraw(FALSE);

  //Get the item which is currently selected
  HTREEITEM hSelItem = GetSelectedItem();
  CString sItem;
  bool bExpanded = false;
  if (hSelItem)
  {
    sItem = ItemToPath(hSelItem);
    bExpanded = IsExpanded(hSelItem);
  }

  m_SharedEnumerator.Refresh();

  KillMonitoringThread();

  //Remove all nodes that currently exist
  Clear();

  //Display the folder items in the tree
  if (m_sRootFolder.IsEmpty())
  {
    //Should we insert a desktop node
    CString sDesktopPath;
    if (m_bShowDesktop && (m_Malloc != nullptr))
    {
#pragma warning(suppress: 26403 26409)
      auto pItem = new CFileTreeCtrlItemInfo;
      pItem->m_NodeType = CFileTreeCtrlItemInfo::NodeType::DesktopNode;
      int nIcon = 0;
      int nSelIcon = 0;

      //Get the localized name, path and correct icons for the Desktop
      LPITEMIDLIST lpDPidl = nullptr;
      if (SUCCEEDED(SHGetSpecialFolderLocation(nullptr, CSIDL_DESKTOP, &lpDPidl)))
      {
        SHFILEINFO sfi{};
#pragma warning(suppress: 26490)
        if (SHGetFileInfo(reinterpret_cast<LPCTSTR>(lpDPidl), 0, &sfi, sizeof(sfi), SHGFI_PIDL | SHGFI_DISPLAYNAME))
#pragma warning(suppress: 26485)
          pItem->m_sRelativePath = sfi.szDisplayName;

        //While we are at it, also get the real path of the Desktop
        const HRESULT hr = SHGetFolderPath(nullptr, CSIDL_DESKTOP, nullptr, SHGFP_TYPE_CURRENT, sDesktopPath.GetBuffer(_MAX_PATH));
        sDesktopPath.ReleaseBuffer();
        if (SUCCEEDED(hr))
        {
          pItem->m_sFQPath = sDesktopPath;
          m_sDesktopPath = pItem->m_sFQPath; //Also maintain a copy ourselves, We use this when we are trying to select items in the control
        }

        nIcon = GetIconIndex(lpDPidl);
        nSelIcon = GetSelIconIndex(lpDPidl);

        //Free up the pidl now that we are finished with it
        m_Malloc->Free(lpDPidl);
      }

      //Add it to the tree control
#pragma warning(suppress: 26486)
      m_hRoot = InsertFileItem(TVI_ROOT, pItem, 0, nIcon, nSelIcon, false);
      m_hDesktop = m_hRoot;
    }
    else
      m_hRoot = TVI_ROOT;

    DisplayRootItems();

    //Also show all items underneath the desktop
    #pragma warning(suppress: 6102)
    if (sDesktopPath.GetLength())
      DisplayPath(sDesktopPath, m_hRoot, false, false);
  }
  else
  {
    if (IsUNCServerName(m_sRootFolder))
    {
      //Should we display the root machine
      if (m_bShowRootedFolder)
        m_hRootedFolder = DisplayNetworkComputer(m_hRoot, m_sRootFolder);
      else
        DisplayShares(m_hRoot, m_sRootFolder);
    }
    else
      DisplayPath(m_sRootFolder, m_hRoot, false);
  }

  //Reselect the initially selected item
  if (hSelItem)
    SetSelectedPath(sItem, bExpanded);

  //Turn back on the redraw flag
  SetRedraw(TRUE);
}

HTREEITEM CFileTreeCtrl::InsertFileItem(_In_ HTREEITEM hParent, _In_ CFileTreeCtrlItemInfo* pItem, _In_ int nStateMask, _In_ int nIcon, _In_ int nSelIcon, _In_ bool bCheckForChildren)
{
  //Validate our parameters
  ASSERT(pItem->m_NodeType != CFileTreeCtrlItemInfo::NodeType::UnknownNode);

  //Correct the label if need be
  CString sLabel;
  if (IsDrive(pItem->m_sFQPath) && m_bShowDriveLabels)
    sLabel = GetDriveLabel(pItem->m_sFQPath);
  else
    sLabel = GetCorrectedLabel(pItem);

  //Add the actual item
  TVINSERTSTRUCT tvis{};
  tvis.hParent = hParent;
  tvis.hInsertAfter = TVI_LAST;
  tvis.item.mask = TVIF_CHILDREN | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_TEXT | TVIF_PARAM;
  tvis.item.iImage = nIcon;
  tvis.item.iSelectedImage = nSelIcon;
#pragma warning(suppress: 26490)
  tvis.item.lParam = reinterpret_cast<LPARAM>(pItem);
  tvis.item.pszText = sLabel.GetBuffer(sLabel.GetLength());
  if (bCheckForChildren)
    tvis.item.cChildren = HasGotSubEntries(pItem->m_sFQPath);
  else
    tvis.item.cChildren = TRUE;
  if (nStateMask)
  {
    tvis.item.mask |= TVIF_STATE;
    tvis.item.stateMask |= TVIS_OVERLAYMASK;
    tvis.item.state |= INDEXTOOVERLAYMASK(nStateMask);
  }

  HTREEITEM hItem = InsertItem(&tvis);
  sLabel.ReleaseBuffer();
#pragma warning(suppress: 26487)
  return hItem;
}

bool CFileTreeCtrl::IsExpanded(_In_ HTREEITEM hItem)
{
  TVITEM tvItem{};
  tvItem.hItem = hItem;
  tvItem.mask = TVIF_HANDLE | TVIF_STATE;
  return GetItem(&tvItem) && (tvItem.state & TVIS_EXPANDED);
}

CString CFileTreeCtrl::GetCorrectedLabel(_In_ CFileTreeCtrlItemInfo* pItem)
{
  CString sLabel(pItem->m_sRelativePath);
  switch (m_FileExtensions)
  {
    case HideFileExtension::UseTheShellSetting:
    {
      CString sTempLabel;
      const short nGFT = GetFileTitle(pItem->m_sRelativePath, sTempLabel.GetBuffer(_MAX_PATH), _MAX_PATH);
      sTempLabel.ReleaseBuffer();
      if (IsFile(pItem->m_sFQPath) && (nGFT == 0))
      {
        pItem->m_bExtensionHidden = (sLabel.CompareNoCase(sTempLabel) != 0);
        sLabel = sTempLabel;
      }
      break;
    }
    case HideFileExtension::HideExtension:
    {
      //Remove the extension if the item is a file
      if (IsFile(pItem->m_sFQPath))
      {
        CString sDrive;
        CString sDir;
        CString sFname;
        _tsplitpath_s(pItem->m_sRelativePath, sDrive.GetBuffer(_MAX_DRIVE), _MAX_DRIVE, sDir.GetBuffer(_MAX_DIR), _MAX_DIR, sFname.GetBuffer(_MAX_FNAME), _MAX_FNAME, nullptr, 0);
        sFname.ReleaseBuffer();
        sDir.ReleaseBuffer();
        sDrive.ReleaseBuffer();
        CString sPath;
        _tmakepath_s(sPath.GetBuffer(_MAX_PATH), _MAX_PATH, sDrive, sDir, sFname, nullptr);
        sPath.ReleaseBuffer();
        sLabel = sPath;
        pItem->m_bExtensionHidden = true;
      }
      break;
    }
    default:
    {
      pItem->m_bExtensionHidden = false;
      break;
    }
  }
  return sLabel;
}

void CFileTreeCtrl::DisplayRootItems()
{
  //Add Current User Folder if requested to do so
  if (m_bShowCurrentUserFolder && (m_Malloc != nullptr))
  {
#pragma warning(suppress: 26403 26409)
    auto pItem = new CFileTreeCtrlItemInfo;
    pItem->m_NodeType = CFileTreeCtrlItemInfo::NodeType::CurrentUserFolderNode;
    int nIcon = 0;
    int nSelIcon = 0;

    //Get the localized name and correct icons for the current users folder
    LPITEMIDLIST lpDocsPidl = nullptr;
    if (SUCCEEDED(SHGetSpecialFolderLocation(nullptr, CSIDL_PROFILE, &lpDocsPidl)))
    {
      //Get the display name of the folder
      SHFILEINFO sfi{};
#pragma warning(suppress: 26490)
      if (SHGetFileInfo(reinterpret_cast<LPCTSTR>(lpDocsPidl), 0, &sfi, sizeof(sfi), SHGFI_PIDL | SHGFI_DISPLAYNAME))
#pragma warning(suppress: 26485)
        pItem->m_sRelativePath = sfi.szDisplayName;

      //While we are at it, also get the real path of the current users folder
      CString sFolderPath;
      const HRESULT hr = SHGetFolderPath(nullptr, CSIDL_PROFILE, nullptr, SHGFP_TYPE_CURRENT, sFolderPath.GetBuffer(_MAX_PATH));
      sFolderPath.ReleaseBuffer();
      if (SUCCEEDED(hr))
      {
        pItem->m_sFQPath = sFolderPath;
        m_sCurrentUserFolderPath = pItem->m_sFQPath; //Also maintain a copy ourselves, We use this when we are trying to select items in the control

        //Get the icons for the folder
        nIcon = GetIconIndex(sFolderPath);
        nSelIcon = GetSelIconIndex(sFolderPath);
      }

      //Free up the pidl now that we are finished with it
      m_Malloc->Free(lpDocsPidl);
    }

    //Add it to the tree control
#pragma warning(suppress: 26486)
    m_hCurrentUserFolderRoot = InsertFileItem(m_hRoot, pItem, 0, nIcon, nSelIcon, false);
  }

  //Add My Documents if requested to do so
  if (m_bShowMyDocuments && (m_Malloc != nullptr))
  {
#pragma warning(suppress: 26403 26409)
    auto pItem = new CFileTreeCtrlItemInfo;
    pItem->m_NodeType = CFileTreeCtrlItemInfo::NodeType::MyDocumentsNode;
    int nIcon = 0;
    int nSelIcon = 0;

    //Get the localized name and correct icons for "My Documents"
    LPITEMIDLIST lpDocsPidl = nullptr;
    if (SUCCEEDED(SHGetSpecialFolderLocation(nullptr, CSIDL_PERSONAL, &lpDocsPidl)))
    {
      //Get the display name of the folder
      SHFILEINFO sfi{};
#pragma warning(suppress: 26490)
      if (SHGetFileInfo(reinterpret_cast<LPCTSTR>(lpDocsPidl), 0, &sfi, sizeof(sfi), SHGFI_PIDL | SHGFI_DISPLAYNAME))
#pragma warning(suppress: 26485)
        pItem->m_sRelativePath = sfi.szDisplayName;

      //While we are at it, also get the real path of "My Documents"
      CString sFolderPath;
      const HRESULT hr = SHGetFolderPath(nullptr, CSIDL_PERSONAL, nullptr, SHGFP_TYPE_CURRENT, sFolderPath.GetBuffer(_MAX_PATH));
      sFolderPath.ReleaseBuffer();
      if (SUCCEEDED(hr))
      {
        pItem->m_sFQPath = sFolderPath;
        m_sMyDocumentsPath = pItem->m_sFQPath; //Also maintain a copy ourselves, We use this when we are trying to select items in the control

        //Get the icons for the folder
        nIcon = GetIconIndex(sFolderPath);
        nSelIcon = GetSelIconIndex(sFolderPath);
      }

      //Free up the pidl now that we are finished with it
      m_Malloc->Free(lpDocsPidl);
    }

    //Add it to the tree control
#pragma warning(suppress: 26486)
    m_hMyDocumentsRoot = InsertFileItem(m_hRoot, pItem, 0, nIcon, nSelIcon, false);
  }

  //Should we insert a "My Computer/This PC" node
  if (m_bShowMyComputer && (m_Malloc != nullptr))
  {
#pragma warning(suppress: 26403 26409)
    auto pItem = new CFileTreeCtrlItemInfo;
    pItem->m_NodeType = CFileTreeCtrlItemInfo::NodeType::MyComputerNode;
    int nIcon = 0;
    int nSelIcon = 0;

    //Get the localized name and correct icons for "My Computer/This PC"
    LPITEMIDLIST lpMCPidl = nullptr;
    if (SUCCEEDED(SHGetSpecialFolderLocation(nullptr, CSIDL_DRIVES, &lpMCPidl)))
    {
      //Get the display name for the folder
      SHFILEINFO sfi{};
#pragma warning(suppress: 26490)
      if (SHGetFileInfo(reinterpret_cast<LPCTSTR>(lpMCPidl), 0, &sfi, sizeof(sfi), SHGFI_PIDL | SHGFI_DISPLAYNAME))
#pragma warning(suppress: 26485)
        pItem->m_sRelativePath = sfi.szDisplayName;

      //Get the icons for the folder
      nIcon = GetIconIndex(lpMCPidl);
      nSelIcon = GetSelIconIndex(lpMCPidl);

      //Free up the pidl now that we are finished with it
      m_Malloc->Free(lpMCPidl);
    }

    //Add it to the tree control
#pragma warning(suppress: 26486)
    m_hMyComputerRoot = InsertFileItem(m_hRoot, pItem, 0, nIcon, nSelIcon, false);
  }

  //Display all the drives (if we are not using My Computer/This PC")
  if (!m_bShowMyComputer)
    DisplayDrives(m_hRoot, false);

  //Also add "Network Neighborhood/Network" if requested to do so
  if (m_bShowNetwork && (m_Malloc != nullptr))
  {
#pragma warning(suppress: 26403 26409)
    auto pItem = new CFileTreeCtrlItemInfo;
    pItem->m_NodeType = CFileTreeCtrlItemInfo::NodeType::NetworkNode;
    int nIcon = 0;
    int nSelIcon = 0;

    //Get the localized name and correct icons for "Network Neighborhood/Network"
    LPITEMIDLIST lpNNPidl = nullptr;
    if (SUCCEEDED(SHGetSpecialFolderLocation(nullptr, CSIDL_NETWORK, &lpNNPidl)))
    {
      //Get the display name for the folder
      SHFILEINFO sfi{};
#pragma warning(suppress: 26490)
      if (SHGetFileInfo(reinterpret_cast<LPCTSTR>(lpNNPidl), 0, &sfi, sizeof(sfi), SHGFI_PIDL | SHGFI_DISPLAYNAME))
#pragma warning(suppress: 26485)
        pItem->m_sRelativePath = sfi.szDisplayName;

      //Get the icons for the folder
      nIcon = GetIconIndex(lpNNPidl);
      nSelIcon = GetSelIconIndex(lpNNPidl);

      //Free up the pidl now that we are finished with it
      m_Malloc->Free(lpNNPidl);
    }

    //Add it to the tree control
#pragma warning(suppress: 26486)
    m_hNetworkRoot = InsertFileItem(m_hRoot, pItem, 0, nIcon, nSelIcon, false);
  }
}

#pragma warning(suppress: 26429)
BOOL CFileTreeCtrl::OnBeginLabelEdit(NMHDR* pNMHDR, LRESULT* pResult) 
{
#pragma warning(suppress: 26490)
  auto pDispInfo = reinterpret_cast<TV_DISPINFO*>(pNMHDR);

  if (pDispInfo->item.hItem)
  {
#pragma warning(suppress: 26486 26490)
    const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(pDispInfo->item.hItem));
#pragma warning(suppress: 26496)
    AFXASSUME(pItem != nullptr);
#pragma warning(suppress: 26489)
    if (m_bAllowRename && ((pItem->m_NodeType == CFileTreeCtrlItemInfo::NodeType::FileNode) || (pItem->m_NodeType == CFileTreeCtrlItemInfo::NodeType::FolderNode)))
      *pResult = FALSE;
    else
      *pResult = TRUE;
  }
  else
    *pResult = TRUE;

  return TRUE; //Allow the message to be reflected again
}

#pragma warning(suppress: 26429)
BOOL CFileTreeCtrl::OnCustomDraw(NMHDR* pNMHDR, LRESULT* pResult) 
{
#pragma warning(suppress: 26490)
  auto pCustomDraw = reinterpret_cast<NMTVCUSTOMDRAW*>(pNMHDR);
  switch (pCustomDraw->nmcd.dwDrawStage) 
  {
    case CDDS_PREPAINT:
    {
      *pResult = CDRF_NOTIFYITEMDRAW; //Tell the control that we are interested in item notifications
      break;
    }
    case CDDS_ITEMPREPAINT:
    {
      //Let it know that we are finished doing custom draw
      *pResult = CDRF_DODEFAULT;

      //To work around what looks like a bug in Commctrl v6 we do not do custom drawing until the window is visible
      if (!IsWindowVisible())
        return TRUE;

      //Check to see if this item is compressed and if it it is, change its color just like explorer does
#pragma warning(suppress: 26490)
      if (m_bShowCompressedUsingDifferentColor && ((pCustomDraw->nmcd.uItemState & CDIS_SELECTED) == 0) && IsCompressed(reinterpret_cast<HTREEITEM>(pCustomDraw->nmcd.dwItemSpec)))
        pCustomDraw->clrText = m_rgbCompressed;
#pragma warning(suppress: 26490)
      else if (m_bShowEncryptedUsingDifferentColor && ((pCustomDraw->nmcd.uItemState & CDIS_SELECTED) == 0) && IsEncrypted(reinterpret_cast<HTREEITEM>(pCustomDraw->nmcd.dwItemSpec))) //also check for encrypted files
        pCustomDraw->clrText = m_rgbEncrypted;

      break;
    }
    default:
    {
      break;
    }
  }

  return TRUE; //Allow the message to be reflected again
}

#pragma warning(suppress: 26429)
BOOL CFileTreeCtrl::OnSelChanged(NMHDR* pNMHDR, LRESULT* pResult)
{
#pragma warning(suppress: 26490)
  auto pNMTreeView = reinterpret_cast<NMTREEVIEW*>(pNMHDR);
#pragma warning(suppress: 26496)
  AFXASSUME(pNMTreeView != nullptr);

  //Nothing selected
  if (pNMTreeView->itemNew.hItem == nullptr)
    return FALSE;

  //Check to see if the current item is valid, if not then delete it (Exclude special items from this check)
#pragma warning(suppress: 26486 26490)
  const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(pNMTreeView->itemNew.hItem));
#pragma warning(suppress: 26496)
  AFXASSUME(pItem != nullptr);
#pragma warning(suppress: 26489)
  CString sPath(pItem->m_sFQPath);
#pragma warning(suppress: 26489)
  if ((pNMTreeView->itemNew.hItem != m_hRoot) &&
      (pNMTreeView->itemNew.hItem != m_hMyDocumentsRoot) &&
      (pNMTreeView->itemNew.hItem != m_hCurrentUserFolderRoot) &&
      (pNMTreeView->itemNew.hItem != m_hNetworkRoot) &&
      (pItem->m_dwNetResourceScope == 0) &&
      (pNMTreeView->itemNew.hItem != m_hMyComputerRoot) && !IsDrive(sPath) && (GetFileAttributes(sPath) == 0xFFFFFFFF))
  {
    //Before we delete it see if we are the only child item
    HTREEITEM hParent = GetParentItem(pNMTreeView->itemNew.hItem);

    //Delete the item
    DeleteItem(pNMTreeView->itemNew.hItem);

    //Remove all the child items from the parent
    SetHasPlusButton(hParent, false);

    *pResult = 1;

    return FALSE; //Allow the message to be reflected again
  }

  //Add to the prev array the item we were just at
  if (pNMTreeView->itemOld.hItem && !m_bUpdatingHistorySelection)
  {
    if (m_PrevItems.size() > m_nMaxHistory)
      m_PrevItems.erase(m_PrevItems.begin());
#pragma warning(suppress: 26486)
    m_PrevItems.push_back(pNMTreeView->itemOld.hItem);
  }

  //Remeber the serial number for this item (if it is a drive)
  if (IsDrive(sPath))
  {
    const int nDrive = sPath.GetAt(0) - _T('A');
#pragma warning(suppress: 26446 26482)
    GetSerialNumber(sPath, m_MediaIDs[nDrive]);
  }

  //Setup notifications if required
  if (m_bAutoRefresh && CanHandleChangeNotifications(sPath))
#pragma warning(suppress: 26489)
    CreateMonitoringThread(pItem->m_sFQPath);
  else
    KillMonitoringThread();

  //call the virtual function
  OnSelectionChanged(pNMTreeView, sPath);

  *pResult = 0;

  return FALSE; //Allow the message to be reflected again
}

#pragma warning(suppress: 26440)
void CFileTreeCtrl::OnSelectionChanged(_In_ NMTREEVIEW* /*pNMTreeView*/, _In_ const CString& /*sPath*/)
{
  //CFileTreeCtrl does nothing, but a derived class may want to do something
}

#pragma warning(suppress: 26429)
BOOL CFileTreeCtrl::OnEndLabelEdit(NMHDR* pNMHDR, LRESULT* pResult) 
{
#pragma warning(suppress: 26490)
  auto pDispInfo = reinterpret_cast<TV_DISPINFO*>(pNMHDR);
#pragma warning(suppress: 26496)
  AFXASSUME(pDispInfo != nullptr);
  if (pDispInfo->item.pszText != nullptr)
  {
    SHFILEOPSTRUCT shfo{};
    shfo.hwnd = GetParent()->GetSafeHwnd();
    shfo.wFunc = FO_RENAME;
    shfo.fFlags = FOF_ALLOWUNDO;

    //Work out the "From" string
    CString sFrom(ItemToPath(pDispInfo->item.hItem));
    const size_t nFromLength = sFrom.GetLength();
    std::vector<TCHAR> sTempFrom;
    sTempFrom.resize(nFromLength + 2);
    _tcscpy_s(sTempFrom.data(), nFromLength + 2, sFrom);
#pragma warning(suppress: 26446)
    sTempFrom[nFromLength + 1] = _T('\0');
    shfo.pFrom = sTempFrom.data();
#pragma warning(suppress: 26486)
    HTREEITEM hParent = GetParentItem(pDispInfo->item.hItem);
    CString sParent(ItemToPath(hParent));

    //Work out the "To" string
    CString sTo;
#pragma warning(suppress: 26486)
    CString sToRelative(pDispInfo->item.pszText);
    if (IsDrive(sParent))
      sTo = sParent + pDispInfo->item.pszText;
    else
      sTo = sParent + _T("\\") + pDispInfo->item.pszText;
#pragma warning(suppress: 26490)
    auto pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(pDispInfo->item.hItem));
#pragma warning(suppress: 26496)
    AFXASSUME(pItem != nullptr);
#pragma warning(suppress: 26489)
    if (pItem->m_bExtensionHidden)
    {
      CString sExt;
      _tsplitpath_s(sFrom, nullptr, 0, nullptr, 0, nullptr, 0, sExt.GetBuffer(_MAX_EXT), _MAX_EXT);
      sExt.ReleaseBuffer();
      sTo += sExt;
      sToRelative += sExt;
    }
    const size_t nToLength = sTo.GetLength();
    std::vector<TCHAR> sTempTo;
    sTempTo.resize(nToLength + 2);
    _tcscpy_s(sTempTo.data(), nToLength + 2, sTo);
#pragma warning(suppress: 26446)
    sTempTo[nToLength + 1] = _T('\0');
    shfo.pTo = sTempTo.data();

    //Kill the monitoring thread if necessary
    const bool bMonitoringThread = (m_pMonitoringThread != nullptr);
    CString sMonitoringPath(m_sMonitoringPath);
    if (bMonitoringThread)
      KillMonitoringThread();

    //Let the shell perform the actual rename
    if ((SHFileOperation(&shfo) == 0) && (shfo.fAnyOperationsAborted == FALSE))
    {
      *pResult = TRUE;

      //Update its text
      SetItemText(pDispInfo->item.hItem, pDispInfo->item.pszText);

      //Update the item data
#pragma warning(suppress: 26489)
      pItem->m_sFQPath = sTo;
#pragma warning(suppress: 26489)
      pItem->m_sRelativePath = sToRelative;

      //Also update the icons for it (if need be)
#pragma warning(suppress: 26489)
      if (!pItem->m_bExtensionHidden)
      {
        CString sPath(ItemToPath(pDispInfo->item.hItem));
        SetItemImage(pDispInfo->item.hItem, GetIconIndex(sPath), GetSelIconIndex(sPath));
      }
    }

    //Recreate the monitoring thread if necessary
    if (bMonitoringThread)
      CreateMonitoringThread(m_sMonitoringPath);
  }
  *pResult = 0;
  return FALSE; //Allow the message to be reflected again
}

BOOL CFileTreeCtrl::PreTranslateMessage(MSG* pMsg)
{
  //When an item is being edited make sure the edit control
  //receives certain important key strokes
  if (GetEditControl())
  {
    ::TranslateMessage(pMsg);
    ::DispatchMessage(pMsg);
    return TRUE; // DO NOT process further
  }

  //Context menu via the keyboard
  if ((((pMsg->message == WM_KEYDOWN || pMsg->message == WM_SYSKEYDOWN) &&
      (pMsg->wParam == VK_F10) && (GetKeyState(VK_SHIFT) & 0x8000)) != 0) ||
      (pMsg->message == WM_CONTEXTMENU))
  {
    CRect rect;
    GetItemRect(GetSelectedItem(), rect, TRUE);
    ClientToScreen(rect);
    OnContextMenu(nullptr, rect.CenterPoint());
    return TRUE;
  }
  //Hitting the Escape key cancels drag & drop
#ifndef FILETREECTRL_NO_RESOURCES
  else if (pMsg->message == WM_KEYDOWN && pMsg->wParam == VK_ESCAPE && IsDragging())
  {
    EndDragging(true);
    return TRUE;
  }
#endif //#ifndef FILETREECTRL_NO_RESOURCES
  //Hitting the Alt-Enter key combination shows the properties sheet 
  else if (pMsg->message == WM_SYSKEYDOWN && pMsg->wParam == VK_RETURN)
  {
    OnProperties();
    return TRUE;
  }
  //Hitting the Enter key opens the item
  else if (pMsg->message == WM_KEYDOWN && pMsg->wParam == VK_RETURN)
  {
    OnOpen();
    return TRUE;
  }
  //Hitting the delete key deletes the item
  else if (pMsg->message == WM_KEYDOWN && pMsg->wParam == VK_DELETE)
  {
    OnDelete();
    return TRUE;
  }
  //hitting the F2 key begins in-place editing of an item
  else if (pMsg->message == WM_KEYDOWN && pMsg->wParam == VK_F2)
  {
    OnRename();
    return TRUE;
  }
  //hitting the F5 key forces a refresh of the whole tree
  else if (pMsg->message == WM_KEYDOWN && pMsg->wParam == VK_F5)
  {
    OnRefresh();
    return TRUE;
  }
  //Hitting the Alt-Left Arrow key combination moves to the previous item
  else if (pMsg->message == WM_SYSKEYDOWN && pMsg->wParam == VK_LEFT)
  {
    OnBack();
    return TRUE;
  }
  //Hitting the Alt-Right Arrow key combination moves to the next item
  else if (pMsg->message == WM_SYSKEYDOWN && pMsg->wParam == VK_RIGHT)
  {
    OnForward();
    return TRUE;
  }
  //hitting the backspace key goes to the parent folder
  else if (pMsg->message == WM_KEYDOWN && pMsg->wParam == VK_BACK)
  {
    UpOneLevel();
    return TRUE;
  }

  //Let the parent class do its thing
  return __super::PreTranslateMessage(pMsg);
}

void CFileTreeCtrl::OnUpOneLevel() 
{
  HTREEITEM hItem = GetSelectedItem();
  if (hItem)
  {
    HTREEITEM hParent = GetParentItem(hItem);
    if (hParent)
      Select(hParent, TVGN_CARET);
  }
}

bool CFileTreeCtrl::IsFile(_In_ HTREEITEM hItem)
{
#pragma warning(suppress: 26490)
  const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(hItem));
#pragma warning(suppress: 26496)
  AFXASSUME(pItem != nullptr);
#pragma warning(suppress: 26489)
  return pItem->m_NodeType == CFileTreeCtrlItemInfo::NodeType::FileNode;
}

bool CFileTreeCtrl::IsFolder(_In_ HTREEITEM hItem)
{
#pragma warning(suppress: 26490)
  const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(hItem));
#pragma warning(suppress: 26496)
  AFXASSUME(pItem != nullptr);
#pragma warning(suppress: 26489)
  return pItem->m_NodeType == CFileTreeCtrlItemInfo::NodeType::FolderNode;
}

bool CFileTreeCtrl::IsDrive(_In_ HTREEITEM hItem)
{
#pragma warning(suppress: 26490)
  const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(hItem));
#pragma warning(suppress: 26496)
  AFXASSUME(pItem != nullptr);
#pragma warning(suppress: 26489)
  return pItem->m_NodeType == CFileTreeCtrlItemInfo::NodeType::DriveNode;
}

bool CFileTreeCtrl::IsNetworkShare(_In_ HTREEITEM hItem)
{
#pragma warning(suppress: 26490)
  const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(hItem));
#pragma warning(suppress: 26496)
  AFXASSUME(pItem != nullptr);

#pragma warning(suppress: 26489)
  return ((pItem->m_NodeType == CFileTreeCtrlItemInfo::NodeType::NetworkNode) && (pItem->m_dwNetResourceDisplayType == RESOURCEDISPLAYTYPE_SHARE));
}

bool CFileTreeCtrl::DriveHasRemovableMedia(_In_ const CString& sPath)
{
  bool bRemovableMedia = false;
  if (IsDrive(sPath))
  {
    const UINT nDriveType = GetDriveType(sPath);
    bRemovableMedia = ((nDriveType == DRIVE_REMOVABLE) || (nDriveType == DRIVE_CDROM));
  }
  return bRemovableMedia;
}

bool CFileTreeCtrl::IsCompressed(_In_ HTREEITEM hItem)
{
  return IsCompressed(ItemToPath(hItem));
}

bool CFileTreeCtrl::IsEncrypted(_In_ HTREEITEM hItem)
{
  return IsEncrypted(ItemToPath(hItem));
}

bool CFileTreeCtrl::IsFile(_In_ const CString& sPath) noexcept
{
  const DWORD dwAttributes = GetFileAttributes(sPath);
  return ((dwAttributes != 0xFFFFFFFF) && ((dwAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0));
}

bool CFileTreeCtrl::IsFolder(_In_ const CString& sPath) noexcept
{
  const DWORD dwAttributes = GetFileAttributes(sPath);
  return ((dwAttributes != 0xFFFFFFFF) && (dwAttributes & FILE_ATTRIBUTE_DIRECTORY));
}

bool CFileTreeCtrl::IsDrive(_In_ const CString& sPath)
{
  return (sPath.GetLength() == 3 && sPath.GetAt(1) == _T(':') && sPath.GetAt(2) == _T('\\'));
}

bool CFileTreeCtrl::IsUNCServerName(_In_ const CString& sPath)
{
  CString sDrive;
  CString sDir;
  CString sFname;
  CString sExt;
  _tsplitpath_s(sPath, sDrive.GetBuffer(_MAX_DRIVE), _MAX_DRIVE, sDir.GetBuffer(_MAX_DIR), _MAX_DIR, sFname.GetBuffer(_MAX_FNAME), _MAX_FNAME, sExt.GetBuffer(_MAX_EXT), _MAX_EXT);
  sExt.ReleaseBuffer();
  sFname.ReleaseBuffer();
  sDir.ReleaseBuffer();
  sDrive.ReleaseBuffer();
  return (sDrive.GetLength() == 0) && (sExt.GetLength() == 0) && (sDir == _T("\\\\")) && (sFname.Find(_T('\\')) == -1);
}

bool CFileTreeCtrl::IsCompressed(_In_ const CString& sPath)
{
  bool bCompressed = false;
  if (!IsDrive(sPath))
  {
    const DWORD dwAttributes = GetFileAttributes(sPath);
    bCompressed = ((dwAttributes != 0xFFFFFFFF) && (dwAttributes & FILE_ATTRIBUTE_COMPRESSED));
  }

  return bCompressed;
}

bool CFileTreeCtrl::IsEncrypted(_In_ const CString& sPath)
{
  bool bEncrypted = false;
  if (!IsDrive(sPath))
  {
    const DWORD dwAttributes = GetFileAttributes(sPath);
    bEncrypted = ((dwAttributes != 0xFFFFFFFF) && (dwAttributes & FILE_ATTRIBUTE_ENCRYPTED));
  }

  return bEncrypted;
}

bool CFileTreeCtrl::HasGotSubEntries(_In_ const CString& sDirectory)
{
  //Validate our parameters
  ASSERT(sDirectory.GetLength());

  if (DriveHasRemovableMedia(sDirectory))
  {
    return true; //we do not bother searching for files on drives which have removable media as this would cause 
                 //the drive to spin up, which for the case of a floppy is annoying
  }
  else
  {
    //First check to see if there is any sub directories
    CFileFind find1;
    CString sFile;
    if (sDirectory.GetAt(sDirectory.GetLength() - 1) == _T('\\'))
      sFile = sDirectory + _T("*.*");
    else
      sFile = sDirectory + _T("\\*.*");
    BOOL bFind = find1.FindFile(sFile);
    while (bFind)
    {
      bFind = find1.FindNextFile();
      if (CanDisplayFolder(find1))
        return true;
    }

    //Now check to see if there is any files of the specified file mask
    for (const auto& sMask : m_FileNameMasks)
    {
      CFileFind find2;
      if (sDirectory.GetAt(sDirectory.GetLength() - 1) == _T('\\'))
        sFile = sDirectory + sMask;
      else
        sFile = sDirectory + _T("\\") + sMask;
      bFind = find2.FindFile(sFile);
      while (bFind)
      {
        bFind = find2.FindNextFile();
        if (CanDisplayFile(find2))
          return true;
      }
    }
  }

  return false;
}

void CFileTreeCtrl::PopulateTree()
{
  //Validate our parameters
  ASSERT(GetSafeHwnd()); //Should only call this function after the creation of it on screen

  //Attach the image list to the tree control
  SetImageList(&m_SystemImageList, TVSIL_NORMAL);

  //Force a refresh
  Refresh();
}

void CFileTreeCtrl::CreateMonitoringThread(_In_ const CString& sPath)
{
  //Kill any current notification thread
  KillMonitoringThread();

  //Setup the monitoring path
  m_sMonitoringPath = sPath;
  const int nLength = m_sMonitoringPath.GetLength();
  ASSERT(nLength);
  if (nLength && m_sMonitoringPath.GetAt(nLength - 1) != _T('\\'))
    m_sMonitoringPath += _T('\\');

  TRACE(_T("CFileTreeCtrl::CreateMonitoringThread, Creating monitoring thread for %s\n"), m_sMonitoringPath.GetString());

  ASSERT(m_pMonitoringThread == nullptr);
  m_pMonitoringThread = AfxBeginThread(_MonitoringThread, this, THREAD_PRIORITY_IDLE, 0, CREATE_SUSPENDED);
  if (m_pMonitoringThread != nullptr)
  {
    //We're in charge of deletion of the thread
    m_pMonitoringThread->m_bAutoDelete = FALSE;

    //Resume the thread now that everything is ready to go
    m_pMonitoringThread->ResumeThread();
  }
}

UINT CFileTreeCtrl::_MonitoringThread(_In_ LPVOID pParam)
{
  //Validate our parameters
  ASSERT(pParam != nullptr);
  auto pThis = static_cast<CFileTreeCtrl*>(pParam);
#pragma warning(suppress: 26496)
  AFXASSUME(pThis != nullptr);

  //Convert from the SDK world to the C++ world
  return pThis->MonitoringThread();
}

UINT CFileTreeCtrl::MonitoringThread()
{
  //Form the notification flag to use
  DWORD dwNotifyFilter = FILE_NOTIFY_CHANGE_DIR_NAME;
  if (m_bShowFiles)
    dwNotifyFilter |= FILE_NOTIFY_CHANGE_FILE_NAME;

  //Get a handle to a file change notification object
  HANDLE hChange = FindFirstChangeNotification(m_sMonitoringPath, FALSE, dwNotifyFilter);
  if (hChange != INVALID_HANDLE_VALUE)
  {
    //Setup the handles to wait on
    std::array<HANDLE, 2> handles{ { hChange, m_MonitoringTerminateEvent } };

    //Sleep until a file change notification wakes this thread or m_TerminateEvent becomes
    //set indicating it's time for the thread to end
    bool bContinue = true;
    while (bContinue)
    {
#pragma warning(suppress: 26472)
      const DWORD dwWait = WaitForMultipleObjects(static_cast<DWORD>(handles.size()), handles.data(), FALSE, INFINITE);
      if ((dwWait - WAIT_OBJECT_0) == 0)
      {
        //Respond to the change notification by posting a user defined message back to the GUI thread
        if (!m_bAutoRefresh)
          bContinue = false;
        else
          PostMessage(WM_TREEUPDATE_CHANGE);

        //Move onto the next notification
        FindNextChangeNotification(hChange);
      }
      else
      {
        //Kill the thread
        bContinue = false;
      }
    }

    //Close the handle we have open
    FindCloseChangeNotification(hChange);
  }

  return 0;
}

bool CFileTreeCtrl::GetSerialNumber(_In_ const CString& sDrive, _Out_ DWORD& dwSerialNumber) noexcept
{
  return GetVolumeInformation(sDrive, nullptr, 0, &dwSerialNumber, nullptr, nullptr, nullptr, 0);
}

bool CFileTreeCtrl::IsMediaValid(_In_ const CString& sDrive)
{
  //return true if the drive does not support removable media
  const UINT nDriveType = GetDriveType(sDrive);
  if ((nDriveType != DRIVE_REMOVABLE) && (nDriveType != DRIVE_CDROM))
    return true;

  //Return false if the drive is empty (::GetVolumeInformation fails)
  DWORD dwSerialNumber = 0;
  const int nDrive = sDrive.GetAt(0) - _T('A');
  if (GetSerialNumber(sDrive, dwSerialNumber))
#pragma warning(suppress: 26446 26482)
    m_MediaIDs[nDrive] = dwSerialNumber;
  else
  {
#pragma warning(suppress: 26446 26482)
    m_MediaIDs[nDrive] = 0xFFFFFFFF;
    return false;
  }

  //Also return false if the disk's serial number has changed
#pragma warning(suppress: 26446 26482)
  if ((m_MediaIDs[nDrive] != dwSerialNumber) && (m_MediaIDs[nDrive] != 0xFFFFFFFF))
  {
#pragma warning(suppress: 26446 26482)
    m_MediaIDs[nDrive] = 0xFFFFFFFF;
    return false;
  }

  return true;
}

bool CFileTreeCtrl::EnumNetwork(_In_ HTREEITEM hParent)
{
#pragma warning(suppress: 26490)
  auto pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(hParent));
#pragma warning(suppress: 26496)
  AFXASSUME(pItem != nullptr);

  //Delegate to the other version of this method
  NETRESOURCE netResource{};
#pragma warning(suppress: 26489)
  netResource.dwScope = pItem->m_dwNetResourceScope;
#pragma warning(suppress: 26489)
  netResource.dwType = pItem->m_dwNetResourceType;
#pragma warning(suppress: 26489)
  netResource.dwDisplayType = pItem->m_dwNetResourceDisplayType;
#pragma warning(suppress: 26489)
  netResource.dwUsage = pItem->m_dwNetResourceUsage;
#pragma warning(suppress: 26489)
  netResource.lpLocalName = pItem->m_sNetResourceLocalName.GetBuffer();
#pragma warning(suppress: 26489)
  netResource.lpRemoteName = pItem->m_sNetResourceRemoteName.GetBuffer();
#pragma warning(suppress: 26489)
  netResource.lpComment = pItem->m_sNetResourceComment.GetBuffer();
#pragma warning(suppress: 26489)
  netResource.lpProvider = pItem->m_sNetResourceProvider.GetBuffer();
  const bool bSuccess = EnumNetwork(&netResource, hParent);
#pragma warning(suppress: 26489)
  pItem->m_sNetResourceProvider.ReleaseBuffer();
#pragma warning(suppress: 26489)
  pItem->m_sNetResourceComment.ReleaseBuffer();
#pragma warning(suppress: 26489)
  pItem->m_sNetResourceRemoteName.ReleaseBuffer();
#pragma warning(suppress: 26489)
  pItem->m_sNetResourceLocalName.ReleaseBuffer();
  return bSuccess;
}

bool CFileTreeCtrl::EnumNetwork(_In_opt_ NETRESOURCE* pParentNetResource, _In_ HTREEITEM hParent)
{
  //What will be the return value from this function
  bool bGotChildren = false;

  //Setup for the network enumeration
  HANDLE hEnum = nullptr;
  DWORD dwResult = WNetOpenEnum(pParentNetResource ? RESOURCE_GLOBALNET : RESOURCE_CONTEXT, m_dwNetworkItemTypes, 0, pParentNetResource, &hEnum);

  //Was the read sucessful
  if (dwResult != NO_ERROR)
  {
    TRACE(_T("CFileTreeCtrl::EnumNetwork, Cannot enumerate network, Error:%u\n"), dwResult);
    return false;
  }

  //Do the network enumeration
  DWORD cbBuffer = 16384;
  bool bNeedMoreMemory = true;
  bool bSuccess = false;
  std::vector<BYTE> enumBuffer;
  DWORD cEntries = 0;
  while (bNeedMoreMemory && !bSuccess)
  {
    //Allocate the memory and enumerate
    enumBuffer.resize(cbBuffer);
    cEntries = 0xFFFFFFFF;
    dwResult = WNetEnumResource(hEnum, &cEntries, enumBuffer.data(), &cbBuffer);
    if (dwResult == ERROR_MORE_DATA)
      cbBuffer *= 2;
    else if (dwResult == NO_ERROR)
      bSuccess = true;
    else
      bNeedMoreMemory = false;
  }

  //Enumeration successful?
  if (bSuccess)
  {
    //Scan through the results
#pragma warning(suppress: 26490)
    auto lpnrDrv = reinterpret_cast<const NETRESOURCE*>(enumBuffer.data());
    for (DWORD i=0; i<cEntries; i++)
    {
      //Get the display name for the item
#pragma warning(suppress: 26481 26486)
      CString sDisplayName(lpnrDrv[i].lpRemoteName);
      if (sDisplayName.IsEmpty())
#pragma warning(suppress: 26481 26486)
        sDisplayName = lpnrDrv[i].lpComment;

      //Get the UNC path for the item
      CString sUNCPath;
      if (sDisplayName.Find(_T("\\\\")) == 0)
      {
        sUNCPath = sDisplayName;
        sDisplayName = sDisplayName.Right(sDisplayName.GetLength() - 2);
      }

      //Setup the item data for the new item
#pragma warning(suppress: 26403 26409)
      auto pNewItem = new CFileTreeCtrlItemInfo;
#pragma warning(suppress: 26481 26486)
      pNewItem->m_sNetResourceLocalName = lpnrDrv[i].lpLocalName;
#pragma warning(suppress: 26481 26486)
      pNewItem->m_sNetResourceRemoteName = lpnrDrv[i].lpRemoteName;
#pragma warning(suppress: 26481 26486)
      pNewItem->m_sNetResourceComment = lpnrDrv[i].lpComment;
#pragma warning(suppress: 26481 26486)
      pNewItem->m_sNetResourceProvider = lpnrDrv[i].lpProvider;
      if (sUNCPath.GetLength())
        pNewItem->m_sFQPath = sUNCPath;
      pNewItem->m_sRelativePath = sDisplayName;
      pNewItem->m_NodeType = CFileTreeCtrlItemInfo::NodeType::NetworkNode;
#pragma warning(suppress: 26481)
      pNewItem->m_dwNetResourceScope = lpnrDrv[i].dwScope;
#pragma warning(suppress: 26481)
      pNewItem->m_dwNetResourceType = lpnrDrv[i].dwType;
#pragma warning(suppress: 26481)
      pNewItem->m_dwNetResourceDisplayType = lpnrDrv[i].dwDisplayType;
#pragma warning(suppress: 26481)
      pNewItem->m_dwNetResourceUsage = lpnrDrv[i].dwUsage;

      //Display a share and the appropriate icon
#pragma warning(suppress: 26481)
      if (lpnrDrv[i].dwDisplayType == RESOURCEDISPLAYTYPE_SHARE)
      {
        //Display only the share name
        const int nPos = pNewItem->m_sRelativePath.Find(_T('\\'));
        if (nPos >= 0)
          pNewItem->m_sRelativePath = pNewItem->m_sRelativePath.Mid(nPos+1);

        //Now add the item into the control
#pragma warning(suppress: 26486)
        if (CanDisplayNetworkItem(pNewItem))
          InsertFileItem(hParent, pNewItem, 0, GetIconIndex(pNewItem->m_sFQPath), GetSelIconIndex(pNewItem->m_sFQPath), true);
        else
          delete pNewItem;
      }
#pragma warning(suppress: 26481)
      else if (lpnrDrv[i].dwDisplayType == RESOURCEDISPLAYTYPE_SERVER)
      {
        //Now add the item into the control
#pragma warning(suppress: 26486)
        if (CanDisplayNetworkItem(pNewItem))
        {
          //Note as an optimization we use the cached computer icon indexes for network computers.This helps to
          //speed up the performance of CFileTreeCtrl when browsing domains or workgroups with a large 
          //number of workstations and servers. You can decide to find the icon indexes at runtime if you define
          //the proprocessor value FILETREECTRL_NO_CACHE_COMPUTER_ICON
        #ifndef FILETREECTRL_NO_CACHE_COMPUTER_ICON
          if ((m_nComputerIconIndex != -1) && (m_nSelectedComputerIconIndex != -1))
            InsertFileItem(hParent, pNewItem, 0, m_nComputerIconIndex, m_nSelectedComputerIconIndex, false);
          else
        #endif //#ifndef FILETREECTRL_NO_CACHE_COMPUTER_ICON
          {
            CString sServer;
            sServer.Format(_T("\\\\%s"), pNewItem->m_sRelativePath.GetString());
            InsertFileItem(hParent, pNewItem, 0, GetIconIndex(sServer), GetSelIconIndex(sServer), false);
          }
        }
        else
          delete pNewItem;
      }
      else
      {
        //Now add the item into the control
#pragma warning(suppress: 26486)
        if (CanDisplayNetworkItem(pNewItem))
        {
          //Just use the generic Network Neighborhood/Network icons for everything else
          LPITEMIDLIST lpNNPidl = nullptr;
          int nIcon = 0xFFFF;
          int nSelIcon = nIcon;
          if (SUCCEEDED(SHGetSpecialFolderLocation(nullptr, CSIDL_NETWORK, &lpNNPidl)))
          {
            nIcon = GetIconIndex(lpNNPidl);
            nSelIcon = GetSelIconIndex(lpNNPidl);

            //Free up the pidl now that we are finished with it
            m_Malloc->Free(lpNNPidl);
          }

          InsertFileItem(hParent, pNewItem, 0, nIcon, nSelIcon, false);
        }
        else
          delete pNewItem;
      }
      bGotChildren = true;
    }
  }
  else
    TRACE(_T("CFileTreeCtrl::EnumNetwork, Cannot complete network drive enumeration, Error:%u\n"), dwResult);

  //Clean up the enumeration handle
  const DWORD dwReturn = WNetCloseEnum(hEnum);
  UNREFERENCED_PARAMETER(dwReturn);

  //Return whether or not we added any items
  return bGotChildren;
}

HTREEITEM CFileTreeCtrl::DisplayNetworkComputer(_In_ HTREEITEM hParent, _In_ const CString& sComputer)
{
  //Setup the item data for the new item
  #pragma warning(suppress: 26403 26409)
  auto pNewItem = new CFileTreeCtrlItemInfo;
#pragma warning(suppress: 26489)
  pNewItem->m_dwNetResourceScope = RESOURCE_GLOBALNET;
  pNewItem->m_sNetResourceRemoteName = sComputer;
  pNewItem->m_dwNetResourceDisplayType = RESOURCEDISPLAYTYPE_SERVER;
  pNewItem->m_sFQPath = sComputer;
  CString sDisplayName(sComputer);
  if (sDisplayName.Find(_T("\\\\")) == 0)
    sDisplayName = sDisplayName.Right(sDisplayName.GetLength() - 2);
  pNewItem->m_sRelativePath = sDisplayName;
  pNewItem->m_NodeType = CFileTreeCtrlItemInfo::NodeType::NetworkNode;

  //What will be the return value from this method (assume the worst)
  HTREEITEM hItem = nullptr;

  //Now add the item into the control
#pragma warning(suppress: 26486)
  if (CanDisplayNetworkItem(pNewItem))
  {
    //Note as an optimization we use the cached computer icon indexes for network computers. This helps to 
    //speed up speed up the performance of CFileTreeCtrl when browsing domains or workgroups with a large 
    //number of workstations and servers. You can decide to find the icon indexes at runtime if you define
    //the proprocessor value FILETREECTRL_NO_CACHE_COMPUTER_ICON
  #ifndef FILETREECTRL_NO_CACHE_COMPUTER_ICON
    if ((m_nComputerIconIndex != -1) && (m_nSelectedComputerIconIndex != -1))
      hItem = InsertFileItem(hParent, pNewItem, 0, m_nComputerIconIndex, m_nSelectedComputerIconIndex, false);
    else
  #endif //#ifndef FILETREECTRL_NO_CACHE_COMPUTER_ICON
    {
      CString sServer;
      sServer.Format(_T("\\\\%s"), pNewItem->m_sRelativePath.GetString());
      hItem = InsertFileItem(hParent, pNewItem, 0, GetIconIndex(sServer), GetSelIconIndex(sServer), false);
    }
  }
  else
    delete pNewItem;

#pragma warning(suppress: 26487)
  return hItem;
}

bool CFileTreeCtrl::DisplayShares(_In_ HTREEITEM hParent, _In_ const CString& sComputer)
{
  //setup a NETRESOURCE which represents "sComputer"
  NETRESOURCE netResource{};
  netResource.dwScope = RESOURCE_GLOBALNET;
  netResource.dwUsage = RESOURCEUSAGE_CONTAINER;
  netResource.dwDisplayType = RESOURCEDISPLAYTYPE_SERVER;
  CString sLocalComputer(sComputer);
  netResource.lpRemoteName = sLocalComputer.GetBuffer();

  //Call the main EnumNetwork method
  const bool bSuccess = EnumNetwork(&netResource, hParent);
  sLocalComputer.ReleaseBuffer();
  return bSuccess;
}

void CFileTreeCtrl::DisplayDrives(_In_ HTREEITEM hParent, _In_ bool bUseSetRedraw)
{
  //Display the wait cursor while we are adding the drives to the control
  CWaitCursor c;

  //Speed up the job by turning off redraw
  if (bUseSetRedraw)
    SetRedraw(FALSE);

  //Enumerate the drive letters and add them to the tree control
  const DWORD dwDrives = GetLogicalDrives();
  DWORD dwMask = 1;
  for (int i=0; i<32; i++)
  {
    if (dwDrives & dwMask)
    {
      CString sDrive;
      sDrive.Format(_T("%c:\\"), i + _T('A'));

      //check if this drive is one of the types to hide
      if (CanDisplayDrive(sDrive))
      {
#pragma warning(suppress: 26403 26409)
        auto pItem = new CFileTreeCtrlItemInfo;
        pItem->m_sFQPath = sDrive;
        pItem->m_sRelativePath = sDrive;
        pItem->m_NodeType = CFileTreeCtrlItemInfo::NodeType::DriveNode;

        //Insert the item into the view
#pragma warning(suppress: 26486)
        InsertFileItem(hParent, pItem, (m_bShowSharedUsingDifferentIcon && IsShared(sDrive)) ? 1 : 0, GetIconIndex(sDrive), GetSelIconIndex(sDrive), true);
      }
    }
    dwMask <<= 1;
  }

  if (bUseSetRedraw)
    SetRedraw(TRUE);
}

CString CFileTreeCtrl::GetDriveLabel(_In_ const CString& sDrive)
{
  //Let's start with the drive letter
  CString sLabel(sDrive);

  //Try to find the item directory using ParseDisplayName
  if ((m_ShellFolder != nullptr) && (m_Malloc != nullptr))
  {
    LPITEMIDLIST lpItem = nullptr;
    HRESULT hr = m_ShellFolder->ParseDisplayName(nullptr, nullptr, CT2W(sDrive), nullptr, &lpItem, nullptr);
    if (SUCCEEDED(hr))
    {
      SHFILEINFO sfi{};
#pragma warning(suppress: 26490)
      if (SHGetFileInfo(reinterpret_cast<LPCTSTR>(lpItem), 0, &sfi, sizeof(sfi), SHGFI_PIDL | SHGFI_DISPLAYNAME))
#pragma warning(suppress: 26485)
        sLabel = sfi.szDisplayName;

      //Free the pidl now that we are finished with it
      m_Malloc->Free(lpItem);
    }
  }

  return sLabel;
}

bool CFileTreeCtrl::IsShared(_In_ const CString& sPath)
{
  //Defer all the work to the share enumerator class
  return m_SharedEnumerator.IsShared(sPath);
}

bool CFileTreeCtrl::IsShortcut(_In_ const CString& sFilename)
{
  CString sExt;
  _tsplitpath_s(sFilename, nullptr, 0, nullptr, 0, nullptr, 0, sExt.GetBuffer(_MAX_EXT), _MAX_EXT);
  sExt.ReleaseBuffer();
  return sExt.CompareNoCase(_T(".LNK")) == 0;
}

void CFileTreeCtrl::DisplayPath(_In_ const CString& sPath, _In_ HTREEITEM hParent, _In_ bool bUseSetRedraw, _In_ bool bDeleteExistingSiblings)
{
  //Display the wait cursor while we are adding the items to the control
  CWaitCursor c;

  //Speed up the job by turning off redraw
  if (bUseSetRedraw)
    SetRedraw(FALSE);

  //Remove all the items currently under hParent if required
  if (bDeleteExistingSiblings)
  {
    HTREEITEM hChild = GetChildItem(hParent);
    while (hChild)
    {
      DeleteItem(hChild);
      hChild = GetChildItem(hParent);
    }
  }

  //Should we display the root folder
  if (m_bShowRootedFolder && m_sRootFolder.GetLength() && (hParent == m_hRoot))
  {
#pragma warning(suppress: 26403 26409)
    auto pItem = new CFileTreeCtrlItemInfo;
    pItem->m_NodeType = CFileTreeCtrlItemInfo::NodeType::FolderNode;
    pItem->m_sFQPath = m_sRootFolder;
    pItem->m_sRelativePath = m_sRootFolder;
#pragma warning(suppress: 26486)
    m_hRootedFolder = InsertFileItem(m_hRoot, pItem, 0, GetIconIndex(m_sRootFolder), GetSelIconIndex(m_sRootFolder), true);
    Expand(m_hRootedFolder, TVE_EXPAND);
    return;
  }

  //Find all the directories underneath sPath
  int nDirectories = 0;
  CFileFind find1;
  CString sFile;
  if (sPath.GetAt(sPath.GetLength() - 1) == _T('\\'))
    sFile = sPath + _T("*.*");
  else
    sFile = sPath + _T("\\*.*");
  BOOL bFind = find1.FindFile(sFile);
  while (bFind)
  {
    bFind = find1.FindNextFile();
    if (find1.IsDirectory())
    {
      if (CanDisplayFolder(find1))
      {
        ++nDirectories;
        CString sTempPath(find1.GetFilePath());

        CString sDrive;
        CString sDir;
        CString sFname;
        CString sExt;
        _tsplitpath_s(sTempPath, sDrive.GetBuffer(_MAX_DRIVE), _MAX_DRIVE, sDir.GetBuffer(_MAX_DIR), _MAX_DIR, sFname.GetBuffer(_MAX_FNAME), _MAX_FNAME, sExt.GetBuffer(_MAX_EXT), _MAX_EXT);
        sExt.ReleaseBuffer();
        sFname.ReleaseBuffer();
        sDir.ReleaseBuffer();
        sDrive.ReleaseBuffer();
        CString sTempPath2;
        _tmakepath_s(sTempPath2.GetBuffer(_MAX_PATH), _MAX_PATH, nullptr, nullptr, sFname, sExt);
        sTempPath2.ReleaseBuffer();

#pragma warning(suppress: 26400 26409)
        auto pItem = new CFileTreeCtrlItemInfo;
        pItem->m_sFQPath = sTempPath;
        pItem->m_sRelativePath = sTempPath2;
        pItem->m_NodeType = CFileTreeCtrlItemInfo::NodeType::FolderNode;

#pragma warning(suppress: 26486)
        InsertFileItem(hParent, pItem, m_bShowSharedUsingDifferentIcon && IsShared(sTempPath) ? 1 : 0, GetIconIndex(sTempPath), GetSelIconIndex(sTempPath), true);
      }
    }
  }

  //find all the files underneath sPath
  int nFiles = 0;
  for (const auto& sMask : m_FileNameMasks)
  {
    CFileFind find2;
    if (sPath.GetAt(sPath.GetLength() - 1) == _T('\\'))
      sFile = sPath + sMask;
    else
      sFile = sPath + _T("\\") + sMask;
    bFind = find2.FindFile(sFile);
    while (bFind)
    {
      bFind = find2.FindNextFile();
      if (!find2.IsDirectory())
      {
        if (CanDisplayFile(find2))
        {
          ++nFiles;
          CString sTempPath(find2.GetFilePath());

          CString sFname;
          CString sExt;
          _tsplitpath_s(sTempPath, nullptr, 0, nullptr, 0, sFname.GetBuffer(_MAX_FNAME), _MAX_FNAME, sExt.GetBuffer(_MAX_EXT), _MAX_EXT);
          sExt.ReleaseBuffer();
          sFname.ReleaseBuffer();
          CString sTempPath2;
          _tmakepath_s(sTempPath2.GetBuffer(_MAX_PATH), _MAX_PATH, nullptr, nullptr, sFname, sExt);
          sTempPath2.ReleaseBuffer();

          auto pItem = new CFileTreeCtrlItemInfo;
          pItem->m_sFQPath = sTempPath;
          pItem->m_sRelativePath = sTempPath2;
          pItem->m_NodeType = CFileTreeCtrlItemInfo::NodeType::FileNode;
          InsertFileItem(hParent, pItem, IsShortcut(sTempPath) ? 2 : 0, GetIconIndex(sTempPath), GetSelIconIndex(sTempPath), true);
        }
      }
    }
  }

  //Now sort the items we just added
  TVSORTCB tvsortcb;
  tvsortcb.hParent = hParent;
  tvsortcb.lpfnCompare = CompareByFilenameNoCase;
  tvsortcb.lParam = 0;
  SortChildrenCB(&tvsortcb);

  //If no items were added then remove the "+" indicator from hParent
  if (nFiles == 0 && nDirectories == 00 && hParent != m_hRoot)
    SetHasPlusButton(hParent, false);

  //Turn back on the redraw flag
  if (bUseSetRedraw)
    SetRedraw(TRUE);
}

void CFileTreeCtrl::DoExpand(_In_ HTREEITEM hItem)
{
#pragma warning(suppress: 26490)
  const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(hItem));
#pragma warning(suppress: 26496)
  AFXASSUME(pItem != nullptr);

  if (hItem == m_hRoot && m_sRootFolder.IsEmpty())
  {
    DisplayRootItems();

    //Also show all items underneath the desktop
    if (m_sDesktopPath.GetLength())
      DisplayPath(m_sDesktopPath, m_hRoot, false, false);
  }
  else if (hItem == m_hMyComputerRoot)
  {
    //Display the wait cursor as this may take some time
    CWaitCursor wait;

    //Enumerate the local drive letters
    DisplayDrives(m_hMyComputerRoot, false);
  }
  else if (hItem == m_hNetworkRoot)
  {
    //Display the wait cursor as this may take some time
    CWaitCursor wait;

    //Enumerate the network resources
    EnumNetwork(nullptr, hItem);
  }
#pragma warning(suppress: 26489)
  else if (pItem->m_dwNetResourceScope && !IsFolder(pItem->m_sFQPath))
  {
    //Display the wait cursor as this may take some time
    CWaitCursor wait;

    //Enumerate the network resources
    EnumNetwork(hItem);
  }
#pragma warning(suppress: 26489)
  else if (pItem->m_sFQPath.GetLength())
  {
#pragma warning(suppress: 26489)
    if (IsMediaValid(pItem->m_sFQPath))
    {
      //Delete the item if the path is no longer valid
#pragma warning(suppress: 26489)
      if (IsFolder(pItem->m_sFQPath))
      {
        //Add the new items to the tree if it does not have any child items already
        if (!GetChildItem(hItem))
#pragma warning(suppress: 26489)
          DisplayPath(pItem->m_sFQPath, hItem);
      }
      else
      {
        //Before we delete it see if we are the only child item
        HTREEITEM hParent = GetParentItem(hItem);

        //Delete the item
        DeleteItem(hItem);

        //Remove all the child items from the parent
        SetHasPlusButton(hParent, false);
      }
    }
    else if (hItem != m_hRoot)
    {
      //Display the wait cursor as this may take some time
      CWaitCursor wait;

      //Collapse the drive node and remove all the child items from it
      Expand(hItem, TVE_COLLAPSE | TVE_COLLAPSERESET);
      SetHasPlusButton(hItem, false);
    }
  }
}

#pragma warning(suppress: 26429)
BOOL CFileTreeCtrl::OnItemExpanding(NMHDR* pNMHDR, LRESULT* pResult)
{
#pragma warning(suppress: 26490)
  auto pNMTreeView = reinterpret_cast<NMTREEVIEW*>(pNMHDR);
#pragma warning(suppress: 26496)
  AFXASSUME(pNMTreeView != nullptr);

  if (pNMTreeView->action == TVE_EXPAND)
  {
    const bool bHasPlus = HasPlusButton(pNMTreeView->itemNew.hItem);
    const bool bHasChildren = (GetChildItem(pNMTreeView->itemNew.hItem) != nullptr);

    if (bHasPlus && !bHasChildren)
      DoExpand(pNMTreeView->itemNew.hItem);
  }
  else if (pNMTreeView->action == TVE_COLLAPSE)
  {
#pragma warning(suppress: 26489 26490)
    const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(pNMTreeView->itemNew.hItem));
#pragma warning(suppress: 26496)
    AFXASSUME(pItem != nullptr);

    //Display the wait cursor  as this may take some time
    CWaitCursor wait;

    //Collapse the node and remove all the child items from it
    const bool bHasChildren = HasPlusButton(pNMTreeView->itemNew.hItem);
    Expand(pNMTreeView->itemNew.hItem, TVE_COLLAPSE | TVE_COLLAPSERESET);

    //Never uppdate the child indicator for a network node which is not a share
    bool bUpdateChildIndicator = true;
#pragma warning(suppress: 26489)
    if (pItem->m_NodeType == CFileTreeCtrlItemInfo::NodeType::NetworkNode)
    {
#pragma warning(suppress: 26489)
      bUpdateChildIndicator = (pItem->m_dwNetResourceDisplayType == RESOURCEDISPLAYTYPE_SHARE);
    }

    if (bUpdateChildIndicator)
      SetHasPlusButton(pNMTreeView->itemNew.hItem, bHasChildren);
  }

  *pResult = 0;

  return FALSE; //Allow the message to be reflected again
}

HTREEITEM CFileTreeCtrl::PathToItem(_In_ const CString& sPath) const
{
  CString sSearch(sPath);
  sSearch.MakeUpper();
  int nSearchLength = sSearch.GetLength();
  if (nSearchLength == 0)
    return nullptr;

  //Remove initial part of path if the root folder is setup
  const int nRootLength = m_sRootFolder.GetLength();
  CString sRootFolder(m_sRootFolder);
  sRootFolder.MakeUpper();
  if (nRootLength)
  {
    if (sSearch.Find(sRootFolder) != 0)
    {
      TRACE(_T("CFileTreeCtrl::PathToItem, Could not find the path %s as the root has been configured as %s\n"), sPath.GetString(), m_sRootFolder.GetString());
      return nullptr;
    }
    sSearch = sSearch.Right(nSearchLength - 1 - nRootLength);
    nSearchLength = sSearch.GetLength();
  }

  //Remove trailing "\" from the path
  if (nSearchLength > 3 && sSearch.GetAt(nSearchLength - 1) == _T('\\'))
    sSearch = sSearch.Left(nSearchLength-1);

  if (sSearch.IsEmpty())
  {
    if (m_hRootedFolder)
      return m_hRootedFolder;
    else
      return nullptr;
  }

  HTREEITEM hItemFound = m_hRoot;
  if (nRootLength && m_hRootedFolder)
    hItemFound = m_hRootedFolder;
  bool bDriveMatch = sRootFolder.IsEmpty();
  const bool bNetworkMatch = m_bShowNetwork && ((sSearch.GetLength() > 2) && sSearch.Find(_T("\\\\")) == 0);
  if (bNetworkMatch)
  {
    bDriveMatch = false;
    hItemFound = FindServersNode(m_hNetworkRoot);
    sSearch = sSearch.Right(sSearch.GetLength() - 2);
  }
  if (bDriveMatch)
  {
    CString sMyDocumentsPath(m_sMyDocumentsPath);
    sMyDocumentsPath.MakeUpper();
    CString sMyDesktopPath(m_sDesktopPath);
    sMyDesktopPath.MakeUpper();
    CString sCurrentUserFolderPath(m_sCurrentUserFolderPath);
    sCurrentUserFolderPath.MakeUpper();

    if (m_hDesktop && sSearch.Find(sMyDesktopPath) == 0)
    {
      hItemFound = m_hDesktop;

      sSearch = sSearch.Right(sSearch.GetLength() - sMyDesktopPath.GetLength() - 1);
      bDriveMatch = false;
    }
    else if (m_hCurrentUserFolderRoot && sSearch.Find(sCurrentUserFolderPath) == 0)
    {
      hItemFound = m_hCurrentUserFolderRoot;

      sSearch = sSearch.Right(sSearch.GetLength() - sCurrentUserFolderPath.GetLength() - 1);
      bDriveMatch = false;
    }
    else if (m_hMyDocumentsRoot && sSearch.Find(sMyDocumentsPath) == 0)
    {
      hItemFound = m_hMyDocumentsRoot;

      sSearch = sSearch.Right(sSearch.GetLength() - sMyDocumentsPath.GetLength() - 1);
      bDriveMatch = false;
    }
    else if (m_hMyComputerRoot)
    {
      hItemFound = m_hMyComputerRoot;
    }
  }

  int nFound = sSearch.Find(_T('\\'));
  while (nFound != -1)
  {
    CString sMatch;
    if (bDriveMatch)
    {
      sMatch = sSearch.Left(nFound + 1);
      bDriveMatch = false;
    }
    else
      sMatch = sSearch.Left(nFound);
    hItemFound = FindSibling(hItemFound, sMatch);
    if (hItemFound == nullptr)
      break;

    sSearch = sSearch.Right(sSearch.GetLength() - nFound - 1);
    nFound = sSearch.Find(_T('\\'));
  };

  //The last item
  if (hItemFound)
  {
    if (sSearch.GetLength())
      hItemFound = FindSibling(hItemFound, sSearch);
  }

  return hItemFound;
}

CString CFileTreeCtrl::ItemToPath(_In_ HTREEITEM hItem) const
{
  CString sPath;
  if (hItem)
  {
#pragma warning(suppress: 26490)
    const CFileTreeCtrlItemInfo* pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(GetItemData(hItem));
#pragma warning(suppress: 26496)
    AFXASSUME(pItem != nullptr);
#pragma warning(suppress: 26489)
    sPath = pItem->m_sFQPath;
  }
  return sPath;
}

#pragma warning(suppress: 26429)
BOOL CFileTreeCtrl::OnRclick(NMHDR* /*pNMHDR*/, LRESULT* pResult)
{
  //Work out the position of where the context menu should be
  const CPoint p(GetCurrentMessage()->pt);
#pragma warning(suppress: 26496)
  CPoint pt(p);
  ScreenToClient(&pt);
  Select(HitTest(pt), TVGN_CARET);
  OnContextMenu(nullptr, p);

  *pResult = 0;

  return FALSE; //Allow the message to be reflected again
}

//Copied from CFrameWnd::OnInitMenuPopup to provide OnUpdateCmdUI functionality in the tree control
#pragma warning(suppress: 26434)
void CFileTreeCtrl::OnInitMenuPopup(CMenu* pMenu, UINT /*nIndex*/, BOOL bSysMenu)
{
  //AfxCancelModes(m_hWnd);

  if (bSysMenu)
    return; //don't support system menu

#pragma warning(suppress: 26496)
  AFXASSUME(pMenu != nullptr);
  //check the enabled state of various menu items

  CCmdUI state;
  state.m_pMenu = pMenu;
  ASSERT(state.m_pOther == nullptr);
  ASSERT(state.m_pParentMenu == nullptr);

  //determine if menu is popup in top-level menu and set m_pOther to
  // it if so (m_pParentMenu == nullptr indicates that it is secondary popup)
  HMENU hParentMenu = nullptr;
  if (AfxGetThreadState()->m_hTrackingMenu == pMenu->m_hMenu)
    state.m_pParentMenu = pMenu; //parent == child for tracking popup
  else if ((hParentMenu = ::GetMenu(m_hWnd)) != nullptr) //NOLINT(clang-analyzer-deadcode.DeadStores)
  {
    CWnd* pParent = GetTopLevelParent();
      //child windows don't have menus -- need to go to the top!
#pragma warning(suppress: 26486)
    if (pParent != nullptr && (hParentMenu = ::GetMenu(pParent->m_hWnd)) != nullptr)
    {
      const int nIndexMax = ::GetMenuItemCount(hParentMenu);
      for (int nIndex = 0; nIndex < nIndexMax; nIndex++)
      {
        if (::GetSubMenu(hParentMenu, nIndex) == pMenu->m_hMenu)
        {
          //when popup is found, m_pParentMenu is containing menu
          state.m_pParentMenu = CMenu::FromHandle(hParentMenu);
          break;
        }
      }
    }
  }

  state.m_nIndexMax = pMenu->GetMenuItemCount();
  for (state.m_nIndex = 0; state.m_nIndex < state.m_nIndexMax; state.m_nIndex++)
  {
    state.m_nID = pMenu->GetMenuItemID(state.m_nIndex);
    if (state.m_nID == 0)
      continue; //menu separator or invalid cmd - ignore it

    ASSERT(state.m_pOther == nullptr);
    ASSERT(state.m_pMenu != nullptr);
#pragma warning(suppress: 26472)
    if (state.m_nID == static_cast<UINT>(-1))
    {
      //possibly a popup menu, route to first item of that popup
      state.m_pSubMenu = pMenu->GetSubMenu(state.m_nIndex);
#pragma warning(suppress: 26472)
      if ((state.m_pSubMenu == nullptr) ||
          ((state.m_nID = state.m_pSubMenu->GetMenuItemID(0)) == 0) ||
          (state.m_nID == static_cast<UINT>(-1)))
      {
        continue; //first item of popup can't be routed to
      }
      state.DoUpdate(this, FALSE); //popups are never auto disabled
    }
    else
    {
      //normal menu item
      //Auto enable/disable if frame window has 'm_bAutoMenuEnable'
      //   set and command is _not_ a system command.
      state.m_pSubMenu = nullptr;
      //state.DoUpdate(this, m_bAutoMenuEnable && state.m_nID < 0xF000);
      state.DoUpdate(this, TRUE && state.m_nID < 0xF000);
    }

    //adjust for menu deletions and additions
    const UINT nCount = pMenu->GetMenuItemCount();
    if (nCount < state.m_nIndexMax)
    {
      state.m_nIndex -= (state.m_nIndexMax - nCount);
      while ((state.m_nIndex < nCount) && (pMenu->GetMenuItemID(state.m_nIndex) == state.m_nID))
      {
        state.m_nIndex++;
      }
    }
    state.m_nIndexMax = nCount;
  }
}

#ifndef FILETREECTRL_NO_RESOURCES
#pragma warning(suppress: 26429)
BOOL CFileTreeCtrl::OnBeginDrag(NMHDR* pNMHDR, LRESULT* pResult) 
{
#pragma warning(suppress: 26490)
  auto pNMTreeView = reinterpret_cast<NMTREEVIEW*>(pNMHDR);
#pragma warning(suppress: 26496)
  AFXASSUME(pNMTreeView != nullptr);

  *pResult = 0;

  if (!m_bAllowDragDrop || !IsDropSource(pNMTreeView->itemNew.hItem)) //NOLINT(clang-analyzer-core.CallAndMessage)
    return FALSE; //Allow the message to be reflected again

  m_pilDrag = CreateDragImage(pNMTreeView->itemNew.hItem);
  if (!m_pilDrag)
    return FALSE; //Allow the message to be reflected again

  m_hItemDrag = pNMTreeView->itemNew.hItem;
  m_hItemDrop = nullptr;

  //Calculate the offset to the hotspot
  CPoint offsetPt(8,8); //Initialize a default offset

  const CPoint dragPt(pNMTreeView->ptDrag); //Get the Drag point
  UINT nHitFlags = 0;
  HTREEITEM hHit = HitTest(dragPt, &nHitFlags);
  if (hHit != nullptr)
  {
    //The drag point has Hit an item in the tree
    CRect itemRect;
    if (GetItemRect(hHit, &itemRect, FALSE))
    {
      //Count indent levels
      HTREEITEM hParent = hHit;
      int nIndentCnt = 0;
      while (hParent != nullptr)
      {
        hParent = GetParentItem(hParent);
        nIndentCnt++;
      }

      if (!(GetStyle() & TVS_LINESATROOT)) 
        nIndentCnt--;

      //Calculate the new offset
      offsetPt.y = dragPt.y - itemRect.top;
      offsetPt.x = dragPt.x - (nIndentCnt * GetIndent()) + GetScrollPos(SB_HORZ);
    }
  }

  //Begin the dragging
  m_pilDrag->BeginDrag(0, offsetPt);
  CPoint pt(pNMTreeView->ptDrag);
  ClientToScreen(&pt);
  m_pilDrag->DragEnter(nullptr, pt);
  SetCapture();

  //Create the timer which is used for auto expansion
  m_nTimerID = SetTimer(FILETREECTRL_TIMERID, 300, nullptr);

  return FALSE; //Allow the message to be reflected again
}

#pragma warning(suppress: 26434)
void CFileTreeCtrl::OnMouseMove(UINT nFlags, CPoint point)
{
  if (IsDragging())
  {
    CRect clientRect;
    GetClientRect(&clientRect);

    //Draw the drag
    CPoint pt(point);
    ClientToScreen(&pt);
    CImageList::DragMove(pt);

    //Only select the drop item if we are in the client area
    HTREEITEM hItem = nullptr;
    if (clientRect.PtInRect(point))
    {
      UINT flags = 0;
      hItem = HitTest(point, &flags);
      if (m_hItemDrop != hItem)
      {
        CImageList::DragShowNolock(FALSE);
        SelectDropTarget(hItem);
        m_hItemDrop = hItem;
        CImageList::DragShowNolock(TRUE);
      }
    }
    
    if (hItem)
      hItem = GetDropTarget(hItem);

    //Change the cursor to give feedback
    if (hItem)
    {
      if ((GetKeyState(VK_CONTROL) & 0x8000))
        SetCursor(m_DropCopyCursor);
      else
        SetCursor(m_DropMoveCursor);
    }
    else
    {
      if ((GetKeyState(VK_CONTROL) & 0x8000))
        SetCursor(m_NoDropCopyCursor);
      else
        SetCursor(m_NoDropMoveCursor);
    }
  }

  //Let the parent class do its thing	
  __super::OnMouseMove(nFlags, point);
}

bool CFileTreeCtrl::IsDropSource(_In_ HTREEITEM hItem)
{
  return IsFile(hItem) || IsFolder(hItem);
}

HTREEITEM CFileTreeCtrl::GetDropTarget(_In_ HTREEITEM hItem)
{
  if (!IsFile(hItem) && (hItem != m_hItemDrag) && (hItem != GetParentItem(m_hItemDrag)) && (IsFolder(hItem) || (IsNetworkShare(hItem))))
  {
    HTREEITEM hParent = hItem;
    while ((hParent = GetParentItem(hParent)) != nullptr)
    {
      if (hParent == m_hItemDrag)
        return nullptr;
    }
    return hItem;
  }
  return nullptr;
}

bool CFileTreeCtrl::IsDragging() noexcept
{
  return (m_pilDrag != nullptr);
}

#pragma warning(suppress: 26434)
void CFileTreeCtrl::OnLButtonUp(UINT nFlags, CPoint point)
{
  CRect clientRect;
  GetClientRect(&clientRect);

  if (clientRect.PtInRect(point))
    EndDragging(false);
  else
    EndDragging(true);

  //Let the parent class do its thing
  __super::OnLButtonUp(nFlags, point);
}

void CFileTreeCtrl::EndDragging(_In_ bool bCancel)
{
  if (IsDragging())
  {
    //Kill the timer that is being used
    KillTimer(m_nTimerID);
    m_nTimerID = 0;

    CImageList::DragLeave(this);
    CImageList::EndDrag();
    ReleaseCapture();

    //Delete the drag image list
    delete m_pilDrag;
    m_pilDrag = nullptr;

    //Remove drop target highlighting
    SelectDropTarget(nullptr);

    //Find out where we are dropping
    m_hItemDrop = GetDropTarget(m_hItemDrop);
    if (m_hItemDrop == nullptr)
      return;

    if (!bCancel)
    {
      //Also need to make the change on disk
      SHFILEOPSTRUCT shfo{};
      shfo.hwnd = GetSafeHwnd();

      if ((GetKeyState(VK_CONTROL) & 0x8000))
        shfo.wFunc = FO_COPY;
      else
        shfo.wFunc = FO_MOVE;

      shfo.fFlags = FOF_SILENT | FOF_NOCONFIRMMKDIR;
      //Undo is not allowed if the SHIFT key is held down
      if (!(GetKeyState(VK_SHIFT) & 0x8000))
        shfo.fFlags |= FOF_ALLOWUNDO;

      CString sFromPath(ItemToPath(m_hItemDrag));
      const size_t nFromLength = sFromPath.GetLength();
      std::vector<TCHAR> sTempFrom;
      sTempFrom.resize(nFromLength + 2);
      _tcscpy_s(sTempFrom.data(), nFromLength + 2, sFromPath);
#pragma warning(suppress: 26446)
      sTempFrom[nFromLength + 1] = _T('\0');
      shfo.pFrom = sTempFrom.data();

      CString sToPath(ItemToPath(m_hItemDrop));
      const size_t nToLength = sToPath.GetLength();
      std::vector<TCHAR> sTempTo;
      sTempTo.resize(nToLength + 2);
      _tcscpy_s(sTempTo.data(), nToLength + 2, sToPath);
#pragma warning(suppress: 26446)
      sTempTo[nToLength + 1] = _T('\0');
      shfo.pTo = sTempTo.data();

      //Kill the monitoring thread if necessary
      const bool bMonitoringThread = (m_pMonitoringThread != nullptr);
      CString sMonitoringPath(m_sMonitoringPath);
      if (bMonitoringThread)
        KillMonitoringThread();

      //Let the shell perform the actual deletion
      const bool bSuccess = ((SHFileOperation(&shfo) == 0) && (shfo.fAnyOperationsAborted == FALSE));

      //Recreate the monitoring thread if necessary
      if (bMonitoringThread)
        CreateMonitoringThread(m_sMonitoringPath);

      if (bSuccess)
      {
        //Only copy the item in the tree if there is not an item with the same
        //text under m_hItemDrop
        CString sText(GetItemText(m_hItemDrag));
        if (!HasChildWithText(m_hItemDrop, sText))
        {
          Expand(m_hItemDrop, TVE_COLLAPSE | TVE_COLLAPSERESET);
          ASSERT(m_hItemDrop != TVI_ROOT);
          SelectItem(m_hItemDrop);

          //Update the children indicator for the folder we just dropped into
          SetHasPlusButton(m_hItemDrop, true);
        }

        if (shfo.wFunc == FO_MOVE)
        {
          //Get the parent of the item we moved prior to deleting it
          HTREEITEM hParent = GetParentItem(m_hItemDrag);

          //Delete the item we just moved
          DeleteItem(m_hItemDrag);

          //Update the children indicator for the item we just dragged from
          const bool bHasChildren = (GetChildItem(hParent) != nullptr);
          if (hParent && !bHasChildren)
            SetHasPlusButton(hParent, false);
        }
      }
    }
  }
}

#pragma warning(suppress: 26434)
void CFileTreeCtrl::OnTimer(UINT_PTR nIDEvent)
{
  if (nIDEvent != m_nTimerID)
  {
    //Let the base class do its thing
    __super::OnTimer(nIDEvent);
    return;
  }

  //Show the dragging effect
  CPoint pt;
  GetCursorPos(&pt);
  RECT rect;
  GetClientRect(&rect);
  ClientToScreen(&rect);
  CImageList::DragMove(pt);

  HTREEITEM hFirstItem = GetFirstVisibleItem();
  CRect ItemRect;
  GetItemRect(hFirstItem, &ItemRect, FALSE);
  if (hFirstItem != GetRootItem() && pt.y < (rect.top + (ItemRect.Height()*2)) && pt.y > rect.top)
  {
    //we need to scroll up
    CImageList::DragShowNolock(FALSE);
    SendMessage(WM_VSCROLL, SB_LINEUP);
    EnsureVisible(hFirstItem);
    SelectDropTarget(hFirstItem);
    m_hItemDrop = hFirstItem;
    CImageList::DragShowNolock(TRUE);
  }
  else if (pt.y > (rect.bottom - (ItemRect.Height()*2)) && pt.y < rect.bottom)
  {
    //we need to scroll down
    CImageList::DragShowNolock(FALSE);
    SendMessage(WM_VSCROLL, SB_LINEDOWN);
    HTREEITEM hLastItem = hFirstItem;
    const int nCount = GetVisibleCount();
    for (int i=0; i<(nCount-1); i++)
      hLastItem = GetNextVisibleItem(hLastItem);
    SelectDropTarget(hLastItem);
    EnsureVisible(hLastItem);
    m_hItemDrop = hLastItem;
    CImageList::DragShowNolock(TRUE);
  }

  //Expand the item if the timer ticks has expired
  if (m_TimerTicks == FILETREECTRL_TIMERTICK_FOR_EXPANSION)
  {
    m_TimerTicks = 0;
    Expand(m_hItemDrop, TVE_EXPAND);
  }

  //Expand the selected item if it is collapsed and
  //the timeout has occurred
  TVITEM tvItem{};
  tvItem.hItem = m_hItemDrop;
  tvItem.mask = TVIF_HANDLE | TVIF_CHILDREN | TVIF_STATE;
  tvItem.stateMask = TVIS_EXPANDED;
#pragma warning(suppress: 26486)
  GetItem(&tvItem);
  if (tvItem.cChildren && ((tvItem.state & TVIS_EXPANDED) == 0))
    m_TimerTicks++;
}
#endif //#ifndef FILETREECTRL_NO_RESOURCES

bool CFileTreeCtrl::HasChildWithText(_In_ HTREEITEM hParent, _In_ const CString& sText)
{
  HTREEITEM hChild = GetChildItem(hParent);
  while (hChild)
  {
    CString sItemText(GetItemText(hChild));
    if (sItemText.CompareNoCase(sText) == 0)
      return true;
    hChild = GetNextSiblingItem(hChild);
  }
  return false;
}

HTREEITEM CFileTreeCtrl::CopyItem(_In_ HTREEITEM hItem, _In_ HTREEITEM hNewParent, _In_ HTREEITEM hAfter)
{
  //Get the details of the item to copy
  TVINSERTSTRUCT tvstruct;
  tvstruct.item.hItem = hItem;
  tvstruct.item.mask = TVIF_CHILDREN | TVIF_HANDLE | TVIF_IMAGE | TVIF_SELECTEDIMAGE | TVIF_PARAM;
  GetItem(&tvstruct.item);
  CString sText(GetItemText(hItem));
  tvstruct.item.cchTextMax = sText.GetLength();
  tvstruct.item.pszText = sText.GetBuffer(tvstruct.item.cchTextMax);

  //Make a copy of the item data we are carying around
#pragma warning(suppress: 26490 26429)
  const CFileTreeCtrlItemInfo* pOldInfo = reinterpret_cast<CFileTreeCtrlItemInfo*>(tvstruct.item.lParam);
#pragma warning(suppress: 26490 26409 26489)
  tvstruct.item.lParam = reinterpret_cast<LPARAM>(new CFileTreeCtrlItemInfo(*pOldInfo));

  //Insert the item at the proper location
  tvstruct.hParent = hNewParent;
  tvstruct.hInsertAfter = hAfter;
  tvstruct.item.mask |= TVIF_TEXT;
  HTREEITEM hNewItem = InsertItem(&tvstruct);

  //Don't forget to release the CString buffer
  sText.ReleaseBuffer();

#pragma warning(suppress: 26487)
  return hNewItem;
}

HTREEITEM CFileTreeCtrl::CopyBranch(_In_ HTREEITEM hBranch, _In_ HTREEITEM hNewParent, _In_ HTREEITEM hAfter)
{
  HTREEITEM hNewItem = CopyItem(hBranch, hNewParent, hAfter);
  HTREEITEM hChild = GetChildItem(hBranch);
  while (hChild != nullptr)
  {
    //recursively transfer all the items
    CopyBranch(hChild, hNewItem);
    hChild = GetNextSiblingItem(hChild);
  }
  return hNewItem;
}

void CFileTreeCtrl::OnBack()
{
  const auto nSize = m_PrevItems.size();
  if (nSize)
  {
    HTREEITEM hOldItem = GetSelectedItem();
#pragma warning(suppress: 26446)
    HTREEITEM hNewItem = m_PrevItems[nSize - 1];

    //Select the previous item
    m_bUpdatingHistorySelection = true;
    m_PrevItems.erase(m_PrevItems.begin() + nSize - 1);
    ASSERT(hNewItem != TVI_ROOT);
#pragma warning(suppress: 26486)
    SelectItem(hNewItem);
    EnsureVisible(hNewItem);
    m_bUpdatingHistorySelection = false;

    //Add the old item to the next stack
    m_NextItems.push_back(hOldItem);
  }
}

#pragma warning(suppress: 26429)
void CFileTreeCtrl::OnUpdateBack(CCmdUI* pCmdUI)
{
#pragma warning(suppress: 26486)
  pCmdUI->Enable(CanGoBack());
}

void CFileTreeCtrl::OnForward()
{
  const auto nSize = m_NextItems.size();
  if (nSize)
  {
    HTREEITEM hOldItem = GetSelectedItem();
#pragma warning(suppress: 26446)
    HTREEITEM hNewItem = m_NextItems[nSize - 1];

    //Select the previous item
    m_bUpdatingHistorySelection = true;
    m_NextItems.erase(m_NextItems.begin() + nSize - 1);
    ASSERT(hNewItem != TVI_ROOT);
#pragma warning(suppress: 26486)
    SelectItem(hNewItem);
    EnsureVisible(hNewItem);
    m_bUpdatingHistorySelection = false;

    //Add the old item to the prev stack
    m_PrevItems.push_back(hOldItem);
  }
}

#pragma warning(suppress: 26429)
void CFileTreeCtrl::OnUpdateForward(CCmdUI* pCmdUI)
{
#pragma warning(suppress: 26486)
  pCmdUI->Enable(CanGoForward());
}

void CFileTreeCtrl::SetMaxHistory(_In_ size_t nMaxHistory)
{
  m_nMaxHistory = nMaxHistory;

  //Shrink the prev array if necessary
  auto nCurItems = m_PrevItems.size();
  if (nCurItems > m_nMaxHistory)
    m_PrevItems.erase(m_PrevItems.begin() + m_nMaxHistory, m_PrevItems.end());

  //Shrink the next array if necessary
  nCurItems = m_NextItems.size();
  if (nCurItems > m_nMaxHistory)
    m_NextItems.erase(m_NextItems.begin() + m_nMaxHistory, m_NextItems.end());
}

size_t CFileTreeCtrl::GetBackSize() const noexcept
{
  return m_PrevItems.size();
}

CString CFileTreeCtrl::GetBackItemText(_In_ size_t nBack) const
{
  //Validate our parameters
  ASSERT(nBack < GetBackSize());

#pragma warning(suppress: 26446)
  HTREEITEM hItem = m_PrevItems[nBack];
  return ItemToPath(hItem);
}

size_t CFileTreeCtrl::GetForwardSize() const noexcept
{
  return m_NextItems.size();
}

CString CFileTreeCtrl::GetForwardItemText(_In_ size_t nForward) const
{
  //Validate our parameters
  ASSERT(nForward < GetForwardSize());

#pragma warning(suppress: 26446)
  HTREEITEM hItem = m_NextItems[nForward];
  return ItemToPath(hItem);
}

void CFileTreeCtrl::KillMonitoringThread()
{
  if (m_pMonitoringThread != nullptr)
  {
    //Kill all the running file change notification thread
    TRACE(_T("CFileTreeCtrl::KillNotificationThread, Killing monitoring thread for %s\n"), m_sMonitoringPath.GetString());

    //Signal the worker thread to exit and wait for it to return
    m_MonitoringTerminateEvent.SetEvent();
    WaitForSingleObject(m_pMonitoringThread->m_hThread, INFINITE);
    delete m_pMonitoringThread;
    m_pMonitoringThread = nullptr;
  }
}

#pragma warning(suppress: 26434)
void CFileTreeCtrl::OnDestroy()
{
  //Kill the monitoring thread
  KillMonitoringThread();

  //Remove all the items from the tree control. This ensures that all the heap memory we have used in the item datas is freed
  Clear();

  //Let the parent class do its thing
  __super::OnDestroy();
}

LRESULT CFileTreeCtrl::OnChange(WPARAM /*wParam*/, LPARAM /*lParam*/)
{
  if (m_bAutoRefresh)
  {
    //Trace message which is helpful for diagnosing autorefresh
    TRACE(_T("CFileTreeCtrl::OnChange, Refreshing %s due to change\n"), m_sMonitoringPath.GetString()),

    Refresh();
  }
  return 0L;
}

void CFileTreeCtrl::SetAutoRefresh(_In_ bool bAutoRefresh)
{
  //Since it can be touched by more than one thead
  m_bAutoRefresh = bAutoRefresh;

  //Force the monitoring threads to be recreated
  Refresh();
}

void CFileTreeCtrl::CollapseExpandBranch(_In_ HTREEITEM hItem, _In_ int nAction)
{
  if (ItemHasChildren(hItem))
  {
    Expand(hItem, nAction);
    hItem = GetChildItem(hItem);
    while (hItem)
    {
      CollapseExpandBranch(hItem, nAction);
      hItem = GetNextSiblingItem(hItem);
    }
  }
}

void CFileTreeCtrl::CollapseAll()
{
  HTREEITEM hItem = GetRootItem();
  while (hItem)
  {
    CollapseExpandBranch(hItem, TVE_COLLAPSE);
    hItem = GetNextSiblingItem(hItem);
  }
}

void CFileTreeCtrl::ExpandAll()
{
  HTREEITEM hItem = GetRootItem();
  while (hItem)
  {
    CollapseExpandBranch(hItem, TVE_EXPAND);
    hItem = GetNextSiblingItem(hItem);
  }
}

void CFileTreeCtrl::Clear()
{
  //Delete all the items
  DeleteAllItems();

  //Reset the member variables we have
  m_hMyComputerRoot = nullptr;
  m_hNetworkRoot = nullptr;
  m_hRootedFolder = nullptr;
  m_hMyDocumentsRoot = nullptr;
  m_hCurrentUserFolderRoot = nullptr;
  m_hDesktop = nullptr;
  m_hRoot = TVI_ROOT;
}

void CFileTreeCtrl::PreSubclassWindow()
{
  //Let the base class do its thing
  __super::PreSubclassWindow();

  //Get a pointer to IShellFolder and IMalloc
  ASSERT(m_ShellFolder == nullptr);
  HRESULT hr = SHGetDesktopFolder(&m_ShellFolder);
  if (FAILED(hr))
  {
    TRACE(_T("CFileTreeCtrl::PreSubclassWindow, Failed in call to SHGetDesktopFolder, Error:%08X\n"), hr);
  }
  ASSERT(m_Malloc == nullptr);
  hr = SHGetMalloc(&m_Malloc);
  if (FAILED(hr))
  {
    TRACE(_T("CFileTreeCtrl::PreSubclassWindow, Failed in call to SHGetMalloc, Error:%08X\n"), hr);
  }

  //Load up the cursors we need
#ifndef FILETREECTRL_NO_RESOURCES
  const CWinApp* pApp = AfxGetApp();
#pragma warning(suppress: 26496)
  AFXASSUME(pApp != nullptr);
  m_NoDropCopyCursor = pApp->LoadCursor(IDR_TREEFILECTRL_NO_DROPCOPY);
  if (m_NoDropCopyCursor == nullptr)
  {
    TRACE(_T("CFileTreeCtrl::PreSubclassWindow, Failed in load IDR_TREEFILECTRL_NO_DROPCOPY cursor\n"));
  }
  m_DropCopyCursor = pApp->LoadCursor(IDR_TREEFILECTRL_DROPCOPY);
  if (m_DropCopyCursor == nullptr)
  {
    TRACE(_T("CFileTreeCtrl::PreSubclassWindow, Failed in load IDR_TREEFILECTRL_DROPCOPY cursor\n"));
  }
  m_NoDropMoveCursor = pApp->LoadCursor(IDR_TREEFILECTRL_NO_DROPMOVE);
  if (m_NoDropMoveCursor == nullptr)
  {
    TRACE(_T("CFileTreeCtrl::PreSubclassWindow, Failed in load IDR_TREEFILECTRL_NO_DROPMOVE cursor\n"));
  }
  m_DropMoveCursor = pApp->LoadStandardCursor(IDC_ARROW);
  if (m_DropMoveCursor == nullptr)
  {
    TRACE(_T("CFileTreeCtrl::PreSubclassWindow, Failed in load IDC_ARROW cursor\n"));
  }
#endif //#ifndef FILETREECTRL_NO_RESOURCES
}

#pragma warning(suppress: 26434 26429 26440)
BOOL CFileTreeCtrl::OnDeleteItem(NMHDR* pNMHDR, LRESULT* pResult)
{
#pragma warning(suppress: 26490)
  auto pNMTreeView = reinterpret_cast<NMTREEVIEW*>(pNMHDR);
#pragma warning(suppress: 26496)
  AFXASSUME(pNMTreeView != nullptr);

  if (pNMTreeView->itemOld.hItem != TVI_ROOT)
  {
#pragma warning(suppress: 26490)
    auto pItem = reinterpret_cast<CFileTreeCtrlItemInfo*>(pNMTreeView->itemOld.lParam);
#pragma warning(suppress: 26486)
    delete pItem;
  }
  *pResult = 0;
  return FALSE; //Allow the message to be reflected again
}

#pragma warning(suppress: 26461)
void DDX_FileTreeValue(_In_ CDataExchange* pDX, _In_ CFileTreeCtrl& ctrlFileTree, _Inout_ CString& sItem)
{
  if (pDX->m_bSaveAndValidate)
    sItem = ctrlFileTree.GetSelectedPath();
  else
    ctrlFileTree.SetSelectedPath(sItem);
}
